# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes.capabilities import (
    UserProfilesAcl,
)
from cognite.client.data_classes.iam import Group
from rich import print
from rich.markup import escape
from rich.prompt import Confirm, Prompt
from rich.table import Table

from .utils import CDFToolConfig


@dataclass
class AuthVariables:
    cluster: str
    project: str
    token: str
    client_id: str
    client_secret: str
    cdf_url: str = None
    token_url: str = None
    tenant_id: str = None
    audience: str = None
    scopes: str = None
    ok: bool = False
    info: str = ""


def get_auth_variables(interactive: bool = False, verbose: bool = False) -> AuthVariables:
    vars = AuthVariables(
        cluster=os.environ.get("CDF_CLUSTER", None),
        project=os.environ.get("CDF_PROJECT", None),
        token=os.environ.get("CDF_TOKEN", None),
        client_id=os.environ.get("IDP_CLIENT_ID", None),
        client_secret=os.environ.get("IDP_CLIENT_SECRET", None),
        cdf_url=os.environ.get("CDF_URL", None),
        tenant_id=os.environ.get("IDP_TENANT_ID", None),
        audience=os.environ.get("IDP_AUDIENCE", None),
        scopes=os.environ.get("IDP_SCOPES", None),
    )
    vars.error = False
    vars.warning = False
    if interactive:
        if vars.cluster is None or len(vars.cluster) == 0:
            vars.cluster = "westeurope-1"
        vars.cluster = Prompt.ask("CDF project cluster (e.g. [italic]westeurope-1[/])? ", default=vars.cluster)
    if vars.cluster is None or len(vars.cluster) == 0:
        vars.error = True
        vars.info += "  [bold red]ERROR[/]: Environment variable CDF_CLUSTER must be set or use --interactive.\n"
        return vars
    if vars.cluster is not None and len(vars.cluster) > 0:
        if verbose:
            vars.info += f"  CDF_CLUSTER={vars.cluster} is set correctly.\n"
    default_cdf_url = f"https://{vars.cluster}.cognitedata.com"
    default_audience = f"https://{vars.cluster}.cognitedata.com"
    default_scopes = f"https://{vars.cluster}.cognitedata.com/.default"
    if vars.cdf_url is None:
        vars.cdf_url = default_cdf_url
    if vars.audience is None:
        vars.audience = default_audience
    if vars.scopes is None:
        vars.scopes = default_scopes
    if interactive:
        vars.cdf_url = Prompt.ask(
            f"What is your CDF URL (recommended: [italic]{default_cdf_url}[/]) ? ", default=vars.cdf_url
        )
        vars.audience = Prompt.ask(
            f"What is your IDP audience (recommended: [italic]{default_audience}[/])? ", default=vars.audience
        )
        vars.scopes = Prompt.ask(
            f"What are your IDP scopes (recommended: [italic]{default_scopes}[/]) ? ", default=vars.scopes
        )
        vars.project = Prompt.ask("CDF project URL name (e.g. [italic]publicdata[/])? ", default=vars.project)
    if vars.cdf_url != f"https://{vars.cluster}.cognitedata.com":
        vars.warning = True
        vars.info += f"  [bold yellow]WARNING[/]: CDF_URL is set to {vars.cdf_url}, are you sure it shouldn't be https://{vars.cluster}.cognitedata.com?\n"
    elif verbose:
        vars.info += "  CDF_URL is set correctly.\n"
    if vars.project is not None and len(vars.project) > 0:
        if verbose:
            vars.info += f"  CDF_PROJECT={vars.project} is set correctly.\n"
    else:
        vars.error = True
        vars.info += "  [bold red]ERROR[/]: Environment variable CDF_PROJECT must be set or use --interactive.\n"
        return vars
    if vars.audience != f"https://{vars.cluster}.cognitedata.com":
        vars.warning = True
        vars.info += f"  [bold yellow]WARNING[/]: IDP_AUDIENCE is set to {vars.audience}, are you sure it shouldn't be https://{vars.cluster}.cognitedata.com?\n"
    elif verbose:
        vars.info += f"  IDP_AUDIENCE = {vars.audience} is set correctly.\n"
    if vars.scopes != f"https://{vars.cluster}.cognitedata.com/.default":
        vars.warning = True
        vars.info += f"  [bold yellow]WARNING[/]: IDP_SCOPES is set to {vars.scopes}, are you sure it shouldn't be https://{vars.cluster}.cognitedata.com/.default?\n"
    elif verbose:
        vars.info += f"  IDP_SCOPES = {vars.scopes} is set correctly.\n"
    if interactive:
        token = False
        if vars.token is None or len(vars.token) == 0:
            token = Confirm.ask(
                "Do you have client id and client secret for a service principal/application? ",
                choices=["y", "n"],
            )
            if not token:
                vars.token = Prompt.ask("OAuth2 token (CDF_TOKEN)? ", password=True)
        else:
            new_token = Prompt.ask(
                "You have set an OAuth2 token (CDF_TOKEN), change it (press ENTER to keep current) ? ",
                password=True,
                default="",
            )
            if len(new_token) > 0:
                vars.token = new_token
            else:
                print("  Keeping existing token.")
        if token:
            azure = Confirm.ask(
                "Do you have Azure Entra/ActiveDirectory as your identity provider ?", choices=["y", "n"]
            )
            name_of_principal = "Service principal/application"
            if azure:
                vars.tenant_id = Prompt.ask(
                    "What is your Entra tenant id (e.g. [italic]12345678-1234-1234-1234-123456789012[/])? ",
                    default=vars.tenant_id,
                )
                name_of_principal = "Application"
                vars.token_url = f"https://login.microsoftonline.com/{vars.tenant_id}/oauth2/v2.0/token"
            else:
                vars.token_url = Prompt.ask(
                    "What is your identity provider token endpoint (e.g. [italic]https://myidp.com/oauth2/token[/])? "
                )
            vars.client_id = Prompt.ask(f"{name_of_principal} client id (CDF_CLIENT_ID)? ", default=vars.client_id)
            if vars.client_secret is not None and len(vars.client_secret) > 0:
                new_secret = Prompt.ask(
                    "You have set a client secret (CDF_CLIENT_SECRET), change it (press ENTER to keep current) ? ",
                    password=True,
                    default="",
                )
                if len(new_secret) > 0:
                    vars.client_secret = new_secret
                else:
                    print("  Keeping existing client secret.")
            else:
                vars.client_secret = Prompt.ask(
                    f"{name_of_principal} client secret (CDF_CLIENT_SECRET)",
                    password=True,
                )
        if (
            vars.client_id is None
            or len(vars.client_id) == 0
            or vars.client_secret is None
            or len(vars.client_secret) == 0
        ):
            if vars.token is None or len(vars.token) == 0:
                vars.error = True
                vars.info += "  [bold red]ERROR[/]: Environment variables IDP_CLIENT_ID and IDP_CLIENT_SECRET (or CDF_TOKEN) must be set.\n"
                return vars
            elif verbose:
                vars.info += "  CDF_TOKEN is set, using it as Bearer token for authorization.\n"
        elif verbose:
            vars.info += "  IDP_CLIENT_ID and IDP_CLIENT_SECRET are set correctly.\n"
    if interactive:
        # Write .env file
        if Path(".env").exists():
            print(
                "[bold yellow]WARNING[/]: .env file already exists and values have been retrieved from it. It will be overwritten."
            )
        write = Confirm.ask(
            "Do you want to save these to .env file for next time ? ",
            choices=["y", "n"],
        )
        if write:
            with open(Path(".env"), "w") as f:
                f.write("# .env file generated by cognite-toolkit\n")
                f.write("CDF_CLUSTER=" + vars.cluster + "\n")
                f.write("CDF_PROJECT=" + vars.project + "\n")
                if vars.token is not None and len(vars.token) > 0:
                    f.write("# When using a token, the IDP variables are not needed, so they are not included.\n")
                    f.write("CDF_TOKEN=" + vars.token + "\n")
                else:
                    f.write("IDP_CLIENT_ID=" + vars.client_id + "\n")
                    f.write("IDP_CLIENT_SECRET=" + vars.client_secret + "\n")
                    if vars.tenant_id is not None and len(vars.tenant_id) > 0:
                        f.write("IDP_TENANT_ID=" + vars.tenant_id + "\n")
                    f.write("IDP_TOKEN_URL=" + vars.token_url + "\n")
                f.write("# The below variables don't have to be set if you have just accepted the defaults.\n")
                f.write("# They are automatically constructed unless they are set.\n")
                f.write("CDF_URL=" + vars.cdf_url + "\n")
                if vars.token is None:
                    f.write("IDP_AUDIENCE=" + vars.audience + "\n")
                    f.write("IDP_SCOPES=" + vars.scopes + "\n")
    return vars


def check_auth(
    ToolGlobals: CDFToolConfig,
    auth_vars: AuthVariables = None,
    group_file: str | None = None,
    update_group: int = 0,
    create_group: str | None = None,
    interactive: bool = False,
    dry_run: bool = False,
    verbose: bool = False,
) -> CogniteClient:
    print("[bold]Checking current service principal/application and environment configurations...[/]")
    if auth_vars is None:
        auth_vars = get_auth_variables(verbose=verbose, interactive=interactive)
    if auth_vars.error:
        print(auth_vars.info)
        ToolGlobals.failed = True
        return
    if auth_vars.warning:
        print(auth_vars.info)
    else:
        print("  [bold green]OK[/]")
    print("Checking basic project configuration...")
    try:
        # Using the token/inspect endpoint to check if the client has access to the project.
        # The response also includes access rights, which can be used to check if the client has the
        # correct access for what you want to do.
        resp = ToolGlobals.client.iam.token.inspect()
        if resp is None or len(resp.capabilities) == 0:
            print(
                "  [bold red]ERROR[/]: Valid authentication token, but it does not give any access rights. Check credentials (CDF_CLIENT_ID/CDF_CLIENT_SECRET or CDF_TOKEN)."
            )
            ToolGlobals.failed = True
            return
        print("  [bold green]OK[/]")
    except Exception:
        print(
            "  [bold red]ERROR[/]: Not a valid authentication token. Check credentials (CDF_CLIENT_ID/CDF_CLIENT_SECRET or CDF_TOKEN)."
        )
        ToolGlobals.failed = True
        return
    try:
        print("Checking projects that the service principal/application has access to...")
        if len(resp.projects) == 0:
            print(
                "  [bold red]ERROR[/]: The service principal/application configured for this client does not have access to any projects."
            )
            ToolGlobals.failed = True
            return
        projects = ""
        projects = projects.join(f"  - {p.url_name}\n" for p in resp.projects)
        print(projects[0:-1])
    except Exception as e:
        print(f"  [bold red]ERROR[/]: Failed to process project information from inspect()\n{e}")
        ToolGlobals.failed = True
        return
    print(f"[italic]Focusing on current project {auth_vars.project} only from here on.[/]")
    print(
        "Checking basic project and group manipulation access rights (projectsAcl: LIST, READ and groupsAcl: LIST, READ, CREATE, UPDATE, DELETE)..."
    )
    try:
        ToolGlobals.verify_client(
            capabilities={
                "projectsAcl": ["LIST", "READ"],
                "groupsAcl": ["LIST", "READ", "CREATE", "UPDATE", "DELETE"],
            }
        )
        print("  [bold green]OK[/]")
    except Exception:
        print(
            "  [bold yellow]WARNING[/]: The service principal/application configured for this client does not have the basic group write access rights."
        )
        print("Checking basic group read access rights (projectsAcl: LIST, READ and groupsAcl: LIST, READ)...")
        try:
            ToolGlobals.verify_client(
                capabilities={
                    "projectsAcl": ["LIST", "READ"],
                    "groupsAcl": ["LIST", "READ"],
                }
            )
            print("  [bold green]OK[/] - can continue with checks.")
        except Exception:
            print(
                "    [bold red]ERROR[/]: Unable to continue, the service principal/application configured for this client does not have the basic read group access rights."
            )
            ToolGlobals.failed = True
            return
    project_info = ToolGlobals.client.get(f"/api/v1/projects/{auth_vars.project}").json()
    print("Checking identity provider settings...")
    oidc = project_info.get("oidcConfiguration", {})
    tenant_id = None
    if "https://login.windows.net" in oidc.get("tokenUrl"):
        tenant_id = oidc.get("tokenUrl").split("/")[-3]
        print(f"  [bold green]OK[/]: Azure Entra (aka ActiveDirectory) with tenant id ({tenant_id}).")
    elif "auth0.com" in oidc.get("tokenUrl"):
        tenant_id = oidc.get("tokenUrl").split("/")[2].split(".")[0]
        print(f"  [bold green]OK[/] - Auth0 with tenant id ({tenant_id}).")
    else:
        print(f"  [bold yellow]WARNING[/]: Unknown identity provider {oidc.get('tokenUrl')}")
    accessClaims = [c.get("claimName") for c in oidc.get("accessClaims", {})]
    print(
        f"  Matching on CDF group sourceIds will be done on any of these claims from the identity provider: {accessClaims}"
    )
    print("Checking CDF group memberships for the current client configured...")
    try:
        groups = ToolGlobals.client.iam.groups.list().data
    except Exception:
        print("  [bold red]ERROR[/]: Unable to retrieve CDF groups.")
        ToolGlobals.failed = True
        return
    read_write = Group.load(
        Path(f"{Path(__file__).parent.parent.as_posix()}{group_file}").read_text(),
    )
    tbl = Table(title="CDF Group ids, Names, and Source Ids")
    tbl.add_column("Id", justify="left")
    tbl.add_column("Name", justify="left")
    tbl.add_column("Source Id", justify="left")
    matched_group_source_id = None
    matched_group_id = 0
    for g in groups:
        if len(groups) > 1 and g.name == read_write.name:
            matched_group_source_id = g.source_id
            matched_group_id = g.id
            tbl.add_row(str(g.id), "[bold]" + g.name + "[/]", g.source_id)
        else:
            tbl.add_row(str(g.id), g.name, g.source_id)
    multiple_groups_with_source_id = 0
    for g in groups:
        if g.source_id == matched_group_source_id:
            multiple_groups_with_source_id += 1
    print(tbl)
    if len(groups) > 1:
        print(
            "  [bold yellow]WARNING[/]: This service principal/application gets its access rights from more than one CDF group."
        )
        print("           This is not recommended. The group matching the group config file is marked in bold above.")
        if update_group == 1:
            print(
                "  [bold red]ERROR[/]: You have specified --update-group=1.\n"
                + "         With multiple groups available, you must use the --update_group=<full-group-i> option to specify which group to update."
            )
            ToolGlobals.failed = True
            return
    else:
        print("  [bold green]OK[/] - Only one group is used for this service principal/application.")
    print("---------------------")
    if matched_group_source_id is not None:
        print("[bold green]RECOMMENDATION[/]:")
        print(f"  You have {multiple_groups_with_source_id} groups with source id {matched_group_source_id},")
        print(
            f"  which is the same source id as the [italic]{escape(read_write.name)}[/] group in the group config file."
        )
        print(
            "  It is recommended that this admin (CI/CD) application/service principal only is member of one group in the identity provider."
        )
        print(
            "  This group's id should be configured as the [italic]readwrite_source_id[/] for the common/cdf_auth_readwrite_all module."
        )
    print(f"\nChecking CDF groups access right against capabilities in {group_file} ...")

    diff = ToolGlobals.client.iam.compare_capabilities(
        resp.capabilities,
        read_write.capabilities,
        project=auth_vars.project,
    )
    if len(diff) > 0:
        for d in diff:
            print(f"  [bold yellow]WARNING[/]: The capability {d} is not present in the CDF project.")
    else:
        print("  [bold green]OK[/] - All capabilities are present in the CDF project.")
    # Flatten out into a list of acls in the existing project
    existing_cap_list = [c.capability for c in resp.capabilities.data]
    if len(groups) > 1:
        print(
            "  [bold yellow]WARNING[/]: This service principal/application gets its access rights from more than one CDF group."
        )
    print("---------------------")
    if len(groups) > 1 and update_group > 1:
        print(f"Checking group config file against capabilities only from the group {update_group}...")
        for g in groups:
            if g.id == update_group:
                existing_cap_list = g.capabilities
                break
    else:
        if len(groups) > 1:
            print("Checking group config file against capabilities across [bold]ALL[/] groups...")
        else:
            print("Checking group config file against capabilities in the group...")

    loosing = ToolGlobals.client.iam.compare_capabilities(
        existing_cap_list,
        resp.capabilities,
        project=auth_vars.project,
    )
    loosing = [l for l in loosing if type(l) is not UserProfilesAcl]  # noqa: E741
    if len(loosing) > 0:
        for d in loosing:
            if len(groups) > 1:
                print(
                    f"  [bold yellow]WARNING[/]: The capability {d} may be lost if\n"
                    + "           switching to relying on only one group based on group config file for access."
                )
            else:
                print(
                    f"  [bold yellow]WARNING[/]: The capability {d} will be removed in the project if overwritten by group config file."
                )
    else:
        print("  [bold green]OK[/] - All capabilities from the CDF project are also present in the group config file.")
    print("---------------------")
    if interactive and matched_group_id != 0:
        push_group = Confirm.ask(
            f"Do you want to update the group with id {matched_group_id} and name {read_write.name} with the capabilities from {group_file} ?",
            choices=["y", "n"],
        )
        if push_group:
            update_group = matched_group_id
    elif interactive:
        push_group = Confirm.ask(
            "Do you want to create a new group with the capabilities from the group config file ?",
            choices=["y", "n"],
        )
        if push_group:
            create_group = Prompt.ask(
                "What is the source id for the new group (typically a group id in the identity provider)? "
            )
    if len(groups) == 1 and update_group == 1:
        update_group = groups[0].id
    elif not interactive and matched_group_id != 0 and update_group == 1:
        update_group = matched_group_id
    if update_group > 1 or create_group is not None:
        if update_group > 0:
            print(f"Updating group {update_group}...")
            for g in groups:
                if g.id == update_group:
                    group = g
                    break
            if group is None:
                print(f"  [bold red]ERROR[/]: Unable to find --group-id={update_group} in CDF.")
                ToolGlobals.failed = True
                return
            read_write.name = group.name
            read_write.source_id = group.source_id
            read_write.metadata = group.metadata
        else:
            print(f"Creating new group based on {group_file}...")
            read_write.source_id = create_group
        try:
            if not dry_run:
                new = ToolGlobals.client.iam.groups.create(read_write)
                print(
                    f"  [bold green]OK[/] - Created new group {new.id} with {len(read_write.capabilities)} capabilities."
                )
            else:
                print(
                    f"  [bold green]OK[/] - Would have created new group with {len(read_write.capabilities)} capabilities."
                )
        except Exception as e:
            print(f"  [bold red]ERROR[/]: Unable to create new group {read_write.name}.\n{e}")
            ToolGlobals.failed = True
            return
        if update_group:
            try:
                if not dry_run:
                    ToolGlobals.client.iam.groups.delete(update_group)
                    print(f"  [bold green]OK[/] - Deleted old group {update_group}.")
                else:
                    print(f"  [bold green]OK[/] - Would have deleted old group {update_group}.")
            except Exception as e:
                print(f"  [bold red]ERROR[/]: Unable to delete old group {update_group}.\n{e}")
                ToolGlobals.failed = True
                return



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import io
import itertools
import json
import re
from abc import ABC, abstractmethod
from collections import Counter, UserList
from collections.abc import Iterable, Sequence, Sized
from contextlib import suppress
from dataclasses import dataclass
from functools import total_ordering
from pathlib import Path
from typing import Any, Generic, Literal, TypeVar, Union, final

import pandas as pd
import yaml
from cognite.client import CogniteClient
from cognite.client.data_classes import (
    DataSet,
    DataSetList,
    ExtractionPipeline,
    ExtractionPipelineConfig,
    ExtractionPipelineList,
    FileMetadata,
    FileMetadataList,
    OidcCredentials,
    TimeSeries,
    TimeSeriesList,
    Transformation,
    TransformationList,
    TransformationSchedule,
    TransformationScheduleList,
    capabilities,
)
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetsAcl,
    ExtractionPipelinesAcl,
    FilesAcl,
    GroupsAcl,
    RawAcl,
    TimeSeriesAcl,
    TransformationsAcl,
)
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerApplyList,
    DataModelApply,
    DataModelApplyList,
    EdgeApply,
    EdgeApplyList,
    NodeApply,
    NodeApplyList,
    SpaceApply,
    SpaceApplyList,
    ViewApply,
    ViewApplyList,
)
from cognite.client.data_classes.data_modeling.ids import ContainerId, DataModelId, EdgeId, NodeId, ViewId
from cognite.client.data_classes.iam import Group, GroupList
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError, CogniteNotFoundError
from rich import print
from rich.table import Table
from typing_extensions import Self

from .delete import delete_instances
from .utils import CDFToolConfig, load_yaml_inject_variables


@dataclass
class RawTable(CogniteObject):
    db_name: str
    table_name: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> RawTable:
        return cls(db_name=resource["dbName"], table_name=resource["tableName"])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "dbName" if camel_case else "db_name": self.db_name,
            "tableName" if camel_case else "table_name": self.table_name,
        }


@dataclass
class LoadableNodes(CogniteObject):
    """
    This is a helper class for nodes that contains arguments that are required for writing the
    nodes to CDF.
    """

    auto_create_direct_relations: bool
    skip_on_version_conflict: bool
    replace: bool
    nodes: NodeApplyList

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return iter(self.nodes)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            auto_create_direct_relations=resource["autoCreateDirectRelations"],
            skip_on_version_conflict=resource["skipOnVersionConflict"],
            replace=resource["replace"],
            nodes=NodeApplyList.load(resource["nodes"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "autoCreateDirectRelations"
            if camel_case
            else "auto_create_direct_relations": self.auto_create_direct_relations,
            "skipOnVersionConflict" if camel_case else "skip_on_version_conflict": self.skip_on_version_conflict,
            "replace": self.replace,
            "nodes": self.nodes.dump(camel_case),
        }


@dataclass
class LoadableEdges(CogniteObject):
    """
    This is a helper class for edges that contains arguments that are required for writing the
    edges to CDF.
    """

    auto_create_start_nodes: bool
    auto_create_end_nodes: bool
    skip_on_version_conflict: bool
    replace: bool
    edges: EdgeApplyList

    def __len__(self):
        return len(self.edges)

    def __iter__(self):
        return iter(self.edges)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            auto_create_start_nodes=resource["autoCreateStartNodes"],
            auto_create_end_nodes=resource["autoCreateEndNodes"],
            skip_on_version_conflict=resource["skipOnVersionConflict"],
            replace=resource["replace"],
            edges=EdgeApplyList.load(resource["edges"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "autoCreateStartNodes" if camel_case else "auto_create_start_nodes": self.auto_create_start_nodes,
            "autoCreateEndNodes" if camel_case else "auto_create_end_nodes": self.auto_create_end_nodes,
            "skipOnVersionConflict" if camel_case else "skip_on_version_conflict": self.skip_on_version_conflict,
            "replace": self.replace,
            "edges": self.edges.dump(camel_case),
        }


@dataclass
class Difference:
    added: list[CogniteResource]
    removed: list[CogniteResource]
    changed: list[CogniteResource]
    unchanged: list[CogniteResource]

    def __iter__(self):
        return iter([self.added, self.removed, self.changed, self.unchanged])

    def __next__(self):
        return next([self.added, self.removed, self.changed, self.unchanged])


T_ID = TypeVar("T_ID", bound=Union[str, int])
T_Resource = TypeVar("T_Resource")
T_ResourceList = TypeVar("T_ResourceList")


class Loader(ABC, Generic[T_ID, T_Resource, T_ResourceList]):
    """
    This is the base class for all loaders. It defines the interface that all loaders must implement.

    A loader is a class that describes how a resource is loaded from a file and uploaded to CDF.

    All resources supported by the cognite_toolkit should implement a loader.

    Class attributes:
        support_drop: Whether the resource supports the drop flag.
        filetypes: The filetypes that are supported by this loader. If empty, all files are supported.
        api_name: The name of the api that is in the cognite_client that is used to interact with the CDF API.
        folder_name: The name of the folder in the build directory where the files are located.
        resource_cls: The class of the resource that is loaded.
        list_cls: The list version of the resource class.
        dependencies: A set of loaders that must be loaded before this loader.
        _display_name: The name of the resource that is used when printing messages. If this is not set the
            api_name is used.
    """

    support_drop = True
    support_upsert = False
    filetypes = frozenset({"yaml", "yml"})
    filename_pattern = ""
    api_name: str
    folder_name: str
    resource_cls: type[CogniteResource]
    list_cls: type[CogniteResourceList]
    identifier_key: str = "externalId"
    dependencies: frozenset[Loader] = frozenset()
    _display_name: str = ""

    def __init__(self, client: CogniteClient, ToolGlobals: CDFToolConfig):
        self.client = client
        self.ToolGlobals = ToolGlobals
        try:
            self.api_class = self._get_api_class(client, self.api_name)
        except AttributeError:
            raise AttributeError(f"Invalid api_name {self.api_name}.")

    @property
    def display_name(self):
        if self._display_name:
            return self._display_name
        return self.api_name

    @staticmethod
    def _get_api_class(client, api_name: str):
        parent = client
        if (dot_count := Counter(api_name)["."]) == 1:
            parent_name, api_name = api_name.split(".")
            parent = getattr(client, parent_name)
        elif dot_count == 0:
            pass
        else:
            raise AttributeError(f"Invalid api_name {api_name}.")
        return getattr(parent, api_name)

    @classmethod
    def create_loader(cls, ToolGlobals: CDFToolConfig):
        client = ToolGlobals.verify_capabilities(capability=cls.get_required_capability(ToolGlobals))
        return cls(client, ToolGlobals)

    @classmethod
    @abstractmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        raise NotImplementedError(f"get_required_capability must be implemented for {cls.__name__}.")

    @classmethod
    @abstractmethod
    def get_id(cls, item: T_Resource) -> T_ID:
        raise NotImplementedError

    @staticmethod
    def fixup_resource(local: T_Resource, remote: T_Resource) -> T_Resource:
        """Takes the local (to be pushed) and remote (from CDF) resource and returns the
        local resource with properties from the remote resource copied over to make
        them equal if we should consider them equal (and skip writing to CDF)."""
        return local

    def remove_unchanged(self, local: T_Resource | Sequence[T_Resource]) -> T_Resource | Sequence[T_Resource]:
        if not isinstance(local, Sequence):
            local = [local]
        if len(local) == 0:
            return local
        try:
            remote = self.retrieve([self.get_id(item) for item in local])
        except CogniteNotFoundError:
            return local
        if len(remote) == 0:
            return local
        for l_resource in local:
            for r in remote:
                if self.get_id(l_resource) == self.get_id(r):
                    r_yaml = self.resource_cls.dump_yaml(r)
                    # To avoid that we mess up the original local resource, we use the
                    # "through yaml copy"-trick to create a copy of the local resource.
                    copy_l = self.resource_cls.load(self.resource_cls.dump_yaml(l_resource))
                    l_yaml = self.resource_cls.dump_yaml(self.fixup_resource(copy_l, r))
                    if l_yaml == r_yaml:
                        local.remove(l_resource)
                        break
        return local

    # Default implementations that can be overridden
    def create(self, items: Sequence[T_Resource], drop: bool, filepath: Path) -> T_ResourceList:
        try:
            created = self.api_class.create(items)
            return created
        except CogniteAPIError as e:
            if e.code == 409:
                print("  [bold yellow]WARNING:[/] Resource(s) already exist(s), skipping creation.")
                return []
            else:
                print(f"[bold red]ERROR:[/] Failed to create resource(s).\n{e}")
                self.ToolGlobals.failed = True
                return []
        except CogniteDuplicatedError as e:
            print(
                f"  [bold yellow]WARNING:[/] {len(e.duplicated)} out of {len(items)} resource(s) already exist(s). {len(e.successful or [])} resource(s) created."
            )
            return []
        except Exception as e:
            print(f"[bold red]ERROR:[/] Failed to create resource(s).\n{e}")
            self.ToolGlobals.failed = True
            return []

    def delete(self, ids: Sequence[T_ID], drop_data: bool) -> int:
        self.api_class.delete(ids)
        return len(ids)

    def retrieve(self, ids: Sequence[T_ID]) -> T_ResourceList:
        return self.api_class.retrieve(ids)

    def load_resource(self, filepath: Path, dry_run: bool) -> T_Resource | T_ResourceList:
        raw_yaml = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        if isinstance(raw_yaml, list):
            return self.list_cls.load(raw_yaml)
        return self.resource_cls.load(raw_yaml)


@final
class AuthLoader(Loader[int, Group, GroupList]):
    support_drop = False
    support_upsert = True
    api_name = "iam.groups"
    folder_name = "auth"
    resource_cls = Group
    list_cls = GroupList
    identifier_key = "name"
    resource_scopes = frozenset(
        {
            capabilities.IDScope,
            capabilities.SpaceIDScope,
            capabilities.DataSetScope,
            capabilities.TableScope,
            capabilities.AssetRootIDScope,
            capabilities.ExtractionPipelineScope,
            capabilities.IDScopeLowerCase,
        }
    )

    def __init__(
        self,
        client: CogniteClient,
        ToolGlobals: CDFToolConfig,
        target_scopes: Literal[
            "all", "all_skipped_validation", "all_scoped_skipped_validation", "resource_scoped_only", "all_scoped_only"
        ] = "all",
    ):
        super().__init__(client, ToolGlobals)
        self.target_scopes = target_scopes

    @property
    def display_name(self):
        if self.target_scopes.startswith("all"):
            scope = "all"
        else:
            scope = "resource scoped"
        return f"{self.api_name}({scope})"

    @staticmethod
    def fixup_resource(local: T_Resource, remote: T_Resource) -> T_Resource:
        local.id = remote.id
        local.is_deleted = False  # If remote is_deleted, this will fail the check.
        local.metadata = remote.metadata  # metadata has no order guarantee, so we exclude it from compare
        local.deleted_time = remote.deleted_time
        return local

    @classmethod
    def create_loader(
        cls,
        ToolGlobals: CDFToolConfig,
        target_scopes: Literal[
            "all", "all_skipped_validation", "all_scoped_skipped_validation", "resource_scoped_only", "all_scoped_only"
        ] = "all",
    ):
        client = ToolGlobals.verify_capabilities(capability=cls.get_required_capability(ToolGlobals))
        return cls(client, ToolGlobals, target_scopes)

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability | list[Capability]:
        return GroupsAcl(
            [GroupsAcl.Action.Read, GroupsAcl.Action.List, GroupsAcl.Action.Create, GroupsAcl.Action.Delete],
            GroupsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: Group) -> str:
        return item.name

    def load_resource(self, filepath: Path, dry_run: bool) -> Group:
        raw = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        for capability in raw.get("capabilities", []):
            for _, values in capability.items():
                if len(values.get("scope", {}).get("datasetScope", {}).get("ids", [])) > 0:
                    if not dry_run and self.target_scopes not in [
                        "all_skipped_validation",
                        "all_scoped_skipped_validation",
                    ]:
                        values["scope"]["datasetScope"]["ids"] = [
                            self.ToolGlobals.verify_dataset(ext_id)
                            for ext_id in values.get("scope", {}).get("datasetScope", {}).get("ids", [])
                        ]
                    else:
                        values["scope"]["datasetScope"]["ids"] = [-1]

                if len(values.get("scope", {}).get("extractionPipelineScope", {}).get("ids", [])) > 0:
                    if not dry_run and self.target_scopes not in [
                        "all_skipped_validation",
                        "all_scoped_skipped_validation",
                    ]:
                        values["scope"]["extractionPipelineScope"]["ids"] = [
                            self.ToolGlobals.verify_extraction_pipeline(ext_id)
                            for ext_id in values.get("scope", {}).get("extractionPipelineScope", {}).get("ids", [])
                        ]
                    else:
                        values["scope"]["extractionPipelineScope"]["ids"] = [-1]
        return Group.load(raw)

    def retrieve(self, ids: Sequence[int]) -> T_ResourceList:
        remote = self.client.iam.groups.list(all=True).data
        found = [g for g in remote if g.name in ids]
        return found

    def delete(self, ids: Sequence[int], drop_data: bool) -> int:
        # Let's prevent that we delete groups we belong to
        try:
            groups = self.client.iam.groups.list().data
        except Exception as e:
            print(
                f"[bold red]ERROR:[/] Failed to retrieve the current service principal's groups. Aborting group deletion.\n{e}"
            )
            return
        my_source_ids = set()
        for g in groups:
            if g.source_id not in my_source_ids:
                my_source_ids.add(g.source_id)
        groups = self.retrieve(ids)
        for g in groups:
            if g.source_id in my_source_ids:
                print(
                    f"  [bold yellow]WARNING:[/] Not deleting group {g.name} with sourceId {g.source_id} as it is used by the current service principal."
                )
                print("     If you want to delete this group, you must do it manually.")
                if g.name not in ids:
                    print(f"    [bold red]ERROR[/] You seem to have duplicate groups of name {g.name}.")
                else:
                    ids.remove(g.name)
        found = [g.id for g in groups if g.name in ids]
        self.client.iam.groups.delete(found)
        return len(found)

    def create(self, items: Sequence[Group], drop: bool, filepath: Path) -> GroupList:
        if self.target_scopes == "all":
            to_create = items
        elif self.target_scopes == "all_skipped_validation":
            raise ValueError("all_skipped_validation is not supported for group creation as scopes would be wrong.")
        elif self.target_scopes == "resource_scoped_only":
            to_create = []
            for item in items:
                item.capabilities = [
                    capability for capability in item.capabilities if type(capability.scope) in self.resource_scopes
                ]
                if item.capabilities:
                    to_create.append(item)
        elif self.target_scopes == "all_scoped_only" or self.target_scopes == "all_scoped_skipped_validation":
            to_create = []
            for item in items:
                item.capabilities = [
                    capability for capability in item.capabilities if type(capability.scope) not in self.resource_scopes
                ]
                if item.capabilities:
                    to_create.append(item)
        else:
            raise ValueError(f"Invalid load value {self.target_scopes}")

        if len(to_create) == 0:
            return []
        # We MUST retrieve all the old groups BEFORE we add the new, if not the new will be deleted
        old_groups = self.client.iam.groups.list(all=True).data
        created = self.client.iam.groups.create(to_create)
        created_names = {g.name for g in created}
        to_delete = [g.id for g in old_groups if g.name in created_names]
        self.client.iam.groups.delete(to_delete)
        return created


@final
class DataSetsLoader(Loader[str, DataSet, DataSetList]):
    support_drop = False
    support_upsert = True
    api_name = "data_sets"
    folder_name = "data_sets"
    resource_cls = DataSet
    list_cls = DataSetList

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        return DataSetsAcl(
            [DataSetsAcl.Action.Read, DataSetsAcl.Action.Write],
            DataSetsAcl.Scope.All(),
        )

    def get_id(self, item: DataSet) -> str:
        return item.external_id

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        raise NotImplementedError("CDF does not support deleting data sets.")

    def retrieve(self, ids: Sequence[str]) -> DataSetList:
        return self.client.data_sets.retrieve_multiple(external_ids=ids)

    @staticmethod
    def fixup_resource(local: DataSet, remote: DataSet) -> DataSet:
        """Sets the read-only properties, id, created_time, and last_updated_time, that are set on the server side.
        This is needed to make the comparison work.
        """

        local.id = remote.id
        local.created_time = remote.created_time
        local.last_updated_time = remote.last_updated_time
        return local

    def load_resource(self, filepath: Path, dry_run: bool) -> DataSetList:
        resource = load_yaml_inject_variables(filepath, {})

        data_sets = [resource] if isinstance(resource, dict) else resource

        for data_set in data_sets:
            if data_set.get("metadata"):
                for key, value in data_set["metadata"].items():
                    data_set["metadata"][key] = json.dumps(value)
        return DataSetList.load(data_sets)

    def create(self, items: Sequence[T_Resource], drop: bool, filepath: Path) -> T_ResourceList | None:
        created = DataSetList([], cognite_client=self.client)
        # There is a bug in the data set API, so only one duplicated data set is returned at the time,
        # so we need to iterate.
        while len(items.data) > 0:
            try:
                created.extend(DataSetList(self.client.data_sets.create(items)))
                return created
            except CogniteDuplicatedError as e:
                if len(e.duplicated) < len(items):
                    for dup in e.duplicated:
                        ext_id = dup.get("externalId", None)
                        for item in items:
                            if item.external_id == ext_id:
                                items.remove(item)
                else:
                    items.data = []
            except Exception as e:
                print(f"[bold red]ERROR:[/] Failed to create data sets.\n{e}")
                self.ToolGlobals.failed = True
                return None
        if len(created) == 0:
            return None
        else:
            return created


@final
class RawLoader(Loader[RawTable, RawTable, list[RawTable]]):
    api_name = "raw.rows"
    folder_name = "raw"
    resource_cls = RawTable
    list_cls = list[RawTable]
    identifier_key = "table_name"
    data_file_types = frozenset({"csv", "parquet"})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        return RawAcl([RawAcl.Action.Read, RawAcl.Action.Write], RawAcl.Scope.All())

    @classmethod
    def get_id(cls, item: RawTable) -> RawTable:
        return item

    def delete(self, ids: Sequence[RawTable], drop_data: bool) -> int:
        count = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            # Raw tables do not have ignore_unknowns_ids, so we need to catch the error
            with suppress(CogniteAPIError):
                tables = [table.table_name for table in raw_tables]
                self.client.raw.tables.delete(db_name=db_name, name=tables)
                count += len(tables)
            if len(self.client.raw.tables.list(db_name=db_name, limit=-1).data) == 0:
                with suppress(CogniteAPIError):
                    self.client.raw.databases.delete(name=db_name)
        return count

    def create(self, items: Sequence[RawTable], drop: bool, filepath: Path) -> list[RawTable]:
        if len(items) != 1:
            raise ValueError("Raw tables must be loaded one at a time.")
        table = items[0]
        datafile = next(
            (
                file
                for file_type in self.data_file_types
                if (file := filepath.parent / f"{table.table_name}.{file_type}").exists()
            ),
            None,
        )
        if datafile is None:
            raise ValueError(f"Failed to find data file for {table.table_name} in {filepath.parent}")
        elif datafile.suffix == ".csv":
            # The replacement is used to ensure that we read exactly the same file on Windows and Linux
            file_content = datafile.read_bytes().replace(b"\r\n", b"\n").decode("utf-8")
            data = pd.read_csv(io.StringIO(file_content), dtype=str)
            data.fillna("", inplace=True)
        elif datafile.suffix == ".parquet":
            data = pd.read_parquet(filepath)
        else:
            raise NotImplementedError(f"Unsupported file type {datafile.suffix} for {table.table_name}")

        self.client.raw.rows.insert_dataframe(
            db_name=table.db_name,
            table_name=table.table_name,
            dataframe=data,
            ensure_parent=True,
        )
        return [table]


@final
class TimeSeriesLoader(Loader[str, TimeSeries, TimeSeriesList]):
    api_name = "time_series"
    folder_name = "timeseries"
    resource_cls = TimeSeriesList
    list_cls = TimeSeriesList
    dependencies = frozenset({DataSetsLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        return TimeSeriesAcl(
            [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
            TimeSeriesAcl.Scope.DataSet([ToolGlobals.data_set_id])
            if ToolGlobals.data_set_id
            else TimeSeriesAcl.Scope.All(),
        )

    def get_id(self, item: TimeSeries) -> str:
        return item.external_id

    def retrieve(self, ids: Sequence[str]) -> TimeSeriesList:
        return self.client.time_series.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        self.client.time_series.delete(external_id=ids, ignore_unknown_ids=True)
        return len(ids)

    def load_resource(self, filepath: Path, dry_run: bool) -> TimeSeries | TimeSeriesList:
        resources = load_yaml_inject_variables(filepath, {})
        if not isinstance(resources, list):
            resources = [resources]
        for resource in resources:
            if resource.get("dataSetExternalId") is not None:
                ds_external_id = resource.pop("dataSetExternalId")
                resource["dataSetId"] = self.ToolGlobals.verify_dataset(ds_external_id) if not dry_run else -1
        return TimeSeriesList.load(resources)


@final
class TransformationLoader(Loader[str, Transformation, TransformationList]):
    api_name = "transformations"
    folder_name = "transformations"
    filename_pattern = (
        r"^(?:(?!\.schedule).)*$"  # Matches all yaml files except file names who's stem contain *.schedule.
    )
    resource_cls = Transformation
    list_cls = TransformationList
    dependencies = frozenset({DataSetsLoader, RawLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        scope = (
            TransformationsAcl.Scope.DataSet([ToolGlobals.data_set_id])
            if ToolGlobals.data_set_id
            else TransformationsAcl.Scope.All()
        )
        return TransformationsAcl(
            [TransformationsAcl.Action.Read, TransformationsAcl.Action.Write],
            scope,
        )

    def get_id(self, item: Transformation) -> str:
        return item.external_id

    def load_resource(self, filepath: Path, dry_run: bool) -> Transformation:
        raw = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        # The `authentication` key is custom for this template:

        source_oidc_credentials = raw.get("authentication", {}).get("read") or raw.get("authentication") or {}
        destination_oidc_credentials = raw.get("authentication", {}).get("write") or raw.get("authentication") or {}
        if raw.get("dataSetExternalId") is not None:
            ds_external_id = raw.pop("dataSetExternalId")
            raw["dataSetId"] = self.ToolGlobals.verify_dataset(ds_external_id) if not dry_run else -1

        transformation = Transformation.load(raw)
        transformation.source_oidc_credentials = source_oidc_credentials and OidcCredentials.load(
            source_oidc_credentials
        )
        transformation.destination_oidc_credentials = destination_oidc_credentials and OidcCredentials.load(
            destination_oidc_credentials
        )
        # Find the non-integer prefixed filename
        file_name = filepath.stem.split(".", 2)[1]
        sql_file = filepath.parent / f"{file_name}.sql"
        if not sql_file.exists():
            sql_file = filepath.parent / f"{transformation.external_id}.sql"
            if not sql_file.exists():
                raise FileNotFoundError(
                    f"Could not find sql file belonging to transformation {filepath.name}. Please run build again."
                )
        transformation.query = sql_file.read_text()
        return transformation

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        self.client.transformations.delete(external_id=ids, ignore_unknown_ids=True)
        return len(ids)

    def create(self, items: Sequence[Transformation], drop: bool, filepath: Path) -> TransformationList:
        try:
            created = self.client.transformations.create(items)
        except CogniteDuplicatedError as e:
            print(
                f"  [bold yellow]WARNING:[/] {len(e.duplicated)} transformation(s) out of {len(items)} transformation(s) already exist(s):"
            )
            for dup in e.duplicated:
                print(f"           {dup.get('externalId', 'N/A')}")
            return []
        except Exception as e:
            print(f"[bold red]ERROR:[/] Failed to create resource(s).\n{e}")
            self.ToolGlobals.failed = True
            return TransformationList([])
        return created


@final
class TransformationScheduleLoader(Loader[str, TransformationSchedule, TransformationScheduleList]):
    api_name = "transformations.schedules"
    folder_name = "transformations"
    filename_pattern = r"^.*\.schedule$"  # Matches all yaml files who's stem contain *.schedule.
    resource_cls = TransformationSchedule
    list_cls = TransformationScheduleList
    dependencies = frozenset({TransformationLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        scope = (
            TransformationsAcl.Scope.DataSet([ToolGlobals.data_set_id])
            if ToolGlobals.data_set_id
            else TransformationsAcl.Scope.All()
        )
        return TransformationsAcl(
            [TransformationsAcl.Action.Read, TransformationsAcl.Action.Write],
            scope,
        )

    def get_id(self, item: Transformation) -> str:
        return item.external_id

    def load_resource(self, filepath: Path, dry_run: bool) -> TransformationSchedule:
        raw = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        return TransformationSchedule.load(raw)

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        try:
            self.client.transformations.schedules.delete(external_id=ids, ignore_unknown_ids=False)
            return len(ids)
        except CogniteNotFoundError as e:
            return len(ids) - len(e.not_found)

    def create(self, items: Sequence[TransformationSchedule], drop: bool, filepath: Path) -> TransformationScheduleList:
        return self.client.transformations.schedules.create(items)


@final
class DatapointsLoader(Loader[list[str], Path, TimeSeriesList]):
    support_drop = False
    filetypes = frozenset({"csv", "parquet"})
    api_name = "time_series.data"
    folder_name = "timeseries_datapoints"
    resource_cls = pd.DataFrame
    dependencies = frozenset({TimeSeriesLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        scope = (
            TimeSeriesAcl.Scope.DataSet([ToolGlobals.data_set_id])
            if ToolGlobals.data_set_id
            else TimeSeriesAcl.Scope.All()
        )

        return TimeSeriesAcl(
            [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
            scope,
        )

    def load_resource(self, filepath: Path, dry_run: bool) -> Path:
        return filepath

    @classmethod
    def get_id(cls, item: Path) -> list[str]:
        raise NotImplementedError

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        # Drop all datapoints?
        raise NotImplementedError()

    def create(self, items: Sequence[Path], drop: bool, filepath: Path) -> TimeSeriesList:
        if len(items) != 1:
            raise ValueError("Datapoints must be loaded one at a time.")
        datafile = items[0]
        if datafile.suffix == ".csv":
            data = pd.read_csv(datafile, parse_dates=True, dayfirst=True, index_col=0)
        elif datafile.suffix == ".parquet":
            data = pd.read_parquet(datafile, engine="pyarrow")
        else:
            raise ValueError(f"Unsupported file type {datafile.suffix} for {datafile.name}")
        self.client.time_series.data.insert_dataframe(data)
        external_ids = [col for col in data.columns if not pd.api.types.is_datetime64_any_dtype(data[col])]
        return TimeSeriesList([TimeSeries(external_id=external_id) for external_id in external_ids])


@final
class ExtractionPipelineLoader(Loader[str, ExtractionPipeline, ExtractionPipelineList]):
    support_drop = True
    api_name = "extraction_pipelines"
    folder_name = "extraction_pipelines"
    filename_pattern = r"^(?:(?!\.config).)*$"  # Matches all yaml files except file names who's stem contain *.config.
    resource_cls = ExtractionPipeline
    list_cls = ExtractionPipelineList
    dependencies = frozenset({DataSetsLoader, RawLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        return ExtractionPipelinesAcl(
            [ExtractionPipelinesAcl.Action.Read, ExtractionPipelinesAcl.Action.Write],
            ExtractionPipelinesAcl.Scope.All(),
        )

    def get_id(self, item: ExtractionPipeline) -> str:
        return item.external_id

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        try:
            self.client.extraction_pipelines.delete(external_id=ids)
            return len(ids)
        except CogniteNotFoundError as e:
            print(
                f"  [bold yellow]WARNING:[/] {len(e.not_found)} out of {len(ids)} extraction pipelines do(es) not exist."
            )

            for dup in e.not_found:
                ext_id = dup.get("externalId", None)
                ids.remove(ext_id)

            if len(ids) > 0:
                self.client.extraction_pipelines.delete(external_id=ids)
                return len(ids)
            return 0

    def load_resource(self, filepath: Path, dry_run: bool) -> ExtractionPipeline:
        resource = load_yaml_inject_variables(filepath, {})

        if resource.get("dataSetExternalId") is not None:
            ds_external_id = resource.pop("dataSetExternalId")
            resource["dataSetId"] = self.ToolGlobals.verify_dataset(ds_external_id) if not dry_run else -1

        return ExtractionPipeline.load(resource)

    def create(self, items: Sequence[ExtractionPipeline], drop: bool, filepath: Path) -> ExtractionPipelineList:
        try:
            return self.client.extraction_pipelines.create(items)
        except CogniteDuplicatedError as e:
            if len(e.duplicated) < len(items):
                for dup in e.duplicated:
                    ext_id = dup.get("externalId", None)
                    for item in items:
                        if item.external_id == ext_id:
                            items.remove(item)
                try:
                    return self.client.extraction_pipelines.create(items)
                except Exception as e:
                    print(f"[bold red]ERROR:[/] Failed to create extraction pipelines.\n{e}")
                    self.ToolGlobals.failed = True
                    return ExtractionPipelineList([])


@final
class ExtractionPipelineConfigLoader(Loader[str, ExtractionPipelineConfig, list[ExtractionPipelineConfig]]):
    support_drop = True
    api_name = "extraction_pipelines.config"
    folder_name = "extraction_pipelines"
    filename_pattern = r"^.*\.config$"
    resource_cls = ExtractionPipelineConfig
    dependencies = frozenset({ExtractionPipelineLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        return ExtractionPipelinesAcl(
            [ExtractionPipelinesAcl.Action.Read, ExtractionPipelinesAcl.Action.Write],
            ExtractionPipelinesAcl.Scope.All(),
        )

    def get_id(self, item: ExtractionPipeline) -> str:
        return item.external_id

    def load_resource(self, filepath: Path, dry_run: bool) -> ExtractionPipelineConfig:
        resource = load_yaml_inject_variables(filepath, {})
        try:
            resource["config"] = yaml.dump(resource.get("config", ""), indent=4)
        except Exception:
            print(
                "[yellow]WARNING:[/] configuration could not be parsed as valid YAML, which is the recommended format.\n"
            )
            resource["config"] = resource.get("config", "")
        return ExtractionPipelineConfig.load(resource)

    def create(
        self, items: Sequence[ExtractionPipelineConfig], drop: bool, filepath: Path
    ) -> list[ExtractionPipelineConfig]:
        try:
            return [self.client.extraction_pipelines.config.create(items[0])]
        except Exception as e:
            print(f"[bold red]ERROR:[/] Failed to create extraction pipelines.\n{e}")
            self.ToolGlobals.failed = True
            return ExtractionPipelineConfig()

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        configs = self.client.extraction_pipelines.config.list(external_id=ids)
        return len(configs)


@final
class FileLoader(Loader[str, FileMetadata, FileMetadataList]):
    api_name = "files"
    filetypes = frozenset({"yaml", "yml"})
    folder_name = "files"
    resource_cls = FileMetadata
    list_cls = FileMetadataList
    dependencies = frozenset({DataSetsLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        if ToolGlobals.data_set_id is None:
            scope = FilesAcl.Scope.All()
        else:
            scope = FilesAcl.Scope.DataSet([ToolGlobals.data_set_id])

        return FilesAcl([FilesAcl.Action.Read, FilesAcl.Action.Write], scope)

    @classmethod
    def get_id(cls, item: FileMetadata) -> str:
        return item.external_id

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        self.client.files.delete(external_id=ids)
        return len(ids)

    def load_resource(self, filepath: Path, dry_run: bool) -> FileMetadata | FileMetadataList:
        try:
            resource = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
            if resource.get("dataSetExternalId") is not None:
                ds_external_id = resource.pop("dataSetExternalId")
                resource["dataSetId"] = self.ToolGlobals.verify_dataset(ds_external_id) if not dry_run else -1
            files = FileMetadataList([FileMetadata.load(resource)])
        except Exception:
            files = FileMetadataList.load(
                load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
            )
        # If we have a file with exact one file config, check to see if this is a pattern to expand
        if len(files.data) == 1 and ("$FILENAME" in files.data[0].external_id or ""):
            # It is, so replace this file with all files in this folder using the same data
            file_data = files.data[0]
            ext_id_pattern = file_data.external_id
            files = FileMetadataList([], cognite_client=self.client)
            for file in filepath.parent.glob("*"):
                if file.suffix[1:] in ["yaml", "yml"]:
                    continue
                files.append(
                    FileMetadata(
                        name=file.name,
                        external_id=re.sub(r"\$FILENAME", file.name, ext_id_pattern),
                        data_set_id=file_data.data_set_id,
                        source=file_data.source,
                        metadata=file_data.metadata,
                        directory=file_data.directory,
                        asset_ids=file_data.asset_ids,
                        labels=file_data.labels,
                        geo_location=file_data.geo_location,
                        security_categories=file_data.security_categories,
                    )
                )
        for file in files.data:
            if not Path(filepath.parent / file.name).exists():
                raise FileNotFoundError(f"Could not find file {file.name} referenced in filepath {filepath.name}")
            if isinstance(file.data_set_id, str):
                # Replace external_id with internal id
                file.data_set_id = self.ToolGlobals.verify_dataset(file.data_set_id) if not dry_run else -1
        return files

    def create(self, items: Sequence[FileMetadata], drop: bool, filepath: Path) -> FileMetadataList:
        created = FileMetadataList([])
        for meta in items:
            datafile = filepath.parent / meta.name
            try:
                created.append(self.client.files.upload(path=datafile, overwrite=drop, **meta.dump(camel_case=False)))
            except CogniteAPIError as e:
                if e.code == 409:
                    print(f"  [bold yellow]WARNING:[/] File {meta.external_id} already exists, skipping upload.")
            except Exception as e:
                print(f"[bold red]ERROR:[/] Failed to upload file {datafile.name}.\n{e}")
                self.ToolGlobals.failed = True
                return created
        return created


@final
class SpaceLoader(Loader[str, SpaceApply, SpaceApplyList]):
    api_name = "data_modeling.spaces"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(space)$"
    resource_cls = SpaceApply
    list_cls = SpaceApplyList
    _display_name = "spaces"

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> list[Capability]:
        return [
            DataModelsAcl(
                [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write],
                DataModelsAcl.Scope.All(),
            ),
            # Needed to delete instances
            DataModelInstancesAcl(
                [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write],
                DataModelInstancesAcl.Scope.All(),
            ),
        ]

    @classmethod
    def get_id(cls, item: SpaceApply) -> str:
        return item.space

    def delete(self, ids: Sequence[str], drop_data: bool) -> int:
        if not drop_data:
            print("  [bold]INFO:[/] Skipping deletion of spaces as drop_data flag is not set...")
            return 0
        print("[bold]Deleting existing data...[/]")
        for space in ids:
            delete_instances(
                ToolGlobals=self.ToolGlobals,
                space_name=space,
            )

        deleted = self.client.data_modeling.spaces.delete(ids)
        return len(deleted)

    def create(self, items: Sequence[SpaceApply], drop: bool, filepath: Path) -> T_ResourceList:
        return self.client.data_modeling.spaces.apply(items)


class ContainerLoader(Loader[ContainerId, ContainerApply, ContainerApplyList]):
    api_name = "data_modeling.containers"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(container)$"
    resource_cls = ContainerApply
    list_cls = ContainerApplyList
    dependencies = frozenset({SpaceLoader})

    _display_name = "containers"

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        # Todo Scoped to spaces
        return DataModelsAcl(
            [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write],
            DataModelsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: ContainerApply) -> ContainerId:
        return item.as_id()

    def delete(self, ids: Sequence[ContainerId], drop_data: bool) -> int:
        if not drop_data:
            print("  [bold]INFO:[/] Skipping deletion of containers as drop_data flag is not set...")
            return 0
        deleted = self.client.data_modeling.containers.delete(ids)
        return len(deleted)

    def create(self, items: Sequence[ContainerApply], drop: bool, filepath: Path) -> T_ResourceList:
        self.ToolGlobals.verify_spaces(list({item.space for item in items}))

        return self.client.data_modeling.containers.apply(items)


class ViewLoader(Loader[ViewId, ViewApply, ViewApplyList]):
    api_name = "data_modeling.views"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(view)$"
    resource_cls = ViewApply
    list_cls = ViewApplyList
    dependencies = frozenset({SpaceLoader, ContainerLoader})

    _display_name = "views"

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        # Todo Scoped to spaces
        return DataModelsAcl(
            [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write],
            DataModelsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: ViewApply) -> ViewId:
        return item.as_id()

    def create(self, items: Sequence[T_Resource], drop: bool, filepath: Path) -> T_ResourceList:
        self.ToolGlobals.verify_spaces(list({item.space for item in items}))
        return self.client.data_modeling.views.apply(items)


@final
class DataModelLoader(Loader[DataModelId, DataModelApply, DataModelApplyList]):
    api_name = "data_modeling.data_models"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(datamodel)$"
    resource_cls = DataModelApply
    list_cls = DataModelApplyList
    dependencies = frozenset({SpaceLoader, ViewLoader})
    _display_name = "data models"

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        # Todo Scoped to spaces
        return DataModelsAcl(
            [DataModelsAcl.Action.Read, DataModelsAcl.Action.Write],
            DataModelsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: DataModelApply) -> DataModelId:
        return item.as_id()

    def create(self, items: Sequence[T_Resource], drop: bool, filepath: Path) -> T_ResourceList:
        self.ToolGlobals.verify_spaces(list({item.space for item in items}))
        return self.client.data_modeling.data_models.apply(items)


@final
class NodeLoader(Loader[list[NodeId], NodeApply, LoadableNodes]):
    api_name = "data_modeling.instances"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(node)$"
    resource_cls = NodeApply
    list_cls = LoadableNodes
    dependencies = frozenset({SpaceLoader, ViewLoader, ContainerLoader})
    _display_name = "nodes"

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        # Todo Scoped to spaces
        return DataModelInstancesAcl(
            [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write],
            DataModelInstancesAcl.Scope.All(),
        )

    def get_id(self, item: NodeApply) -> NodeId:
        return item.as_id()

    def load_resource(self, filepath: Path, dry_run: bool) -> LoadableNodes:
        raw = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        if isinstance(raw, list):
            raise ValueError(f"Unexpected node yaml file format {filepath.name}")
        return LoadableNodes.load(raw, cognite_client=self.client)

    def delete(self, ids: Sequence[NodeId], drop_data: bool) -> int:
        if not drop_data:
            print("  [bold]INFO:[/] Skipping deletion of nodes as drop_data flag is not set...")
            return 0
        deleted = self.client.data_modeling.instances.delete(nodes=ids)
        return len(deleted.nodes)

    def create(self, items: Sequence[LoadableNodes], drop: bool, filepath: Path) -> LoadableNodes:
        if not isinstance(items, LoadableNodes):
            raise ValueError("Unexpected node format file format")
        self.ToolGlobals.verify_spaces(list({item.space for item in items}))
        item = items
        _ = self.client.data_modeling.instances.apply(
            nodes=item.nodes,
            auto_create_direct_relations=item.auto_create_direct_relations,
            skip_on_version_conflict=item.skip_on_version_conflict,
            replace=item.replace,
        )
        return items


@final
class EdgeLoader(Loader[EdgeId, EdgeApply, LoadableEdges]):
    api_name = "data_modeling.instances"
    folder_name = "data_models"
    filename_pattern = r"^.*\.?(edge)$"
    resource_cls = EdgeApply
    list_cls = LoadableEdges
    _display_name = "edges"

    # Note edges do not need nodes to be created first, as they are created as part of the edge creation.
    # However, for deletion (reversed order) we need to delete edges before nodes.
    dependencies = frozenset({SpaceLoader, ViewLoader, NodeLoader})

    @classmethod
    def get_required_capability(cls, ToolGlobals: CDFToolConfig) -> Capability:
        # Todo Scoped to spaces
        return DataModelInstancesAcl(
            [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write],
            DataModelInstancesAcl.Scope.All(),
        )

    def get_id(self, item: EdgeApply) -> EdgeId:
        return item.as_id()

    def load_resource(self, filepath: Path, dry_run: bool) -> LoadableEdges:
        raw = load_yaml_inject_variables(filepath, self.ToolGlobals.environment_variables())
        if isinstance(raw, list):
            raise ValueError(f"Unexpected edge yaml file format {filepath.name}")
        return LoadableEdges.load(raw, cognite_client=self.client)

    def delete(self, ids: Sequence[EdgeId], drop_data: bool) -> int:
        if not drop_data:
            print("  [bold]INFO:[/] Skipping deletion of edges as drop_data flag is not set...")
            return 0
        deleted = self.client.data_modeling.instances.delete(edges=ids)
        return len(deleted.edges)

    def create(self, items: Sequence[LoadableEdges], drop: bool, filepath: Path) -> LoadableEdges:
        if not isinstance(items, LoadableEdges):
            raise ValueError("Unexpected edge format file format")
        self.ToolGlobals.verify_spaces(list({item.space for item in items}))
        item = items
        _ = self.client.data_modeling.instances.apply(
            edges=item.edges,
            auto_create_start_nodes=item.auto_create_start_nodes,
            auto_create_end_nodes=item.auto_create_end_nodes,
            skip_on_version_conflict=item.skip_on_version_conflict,
            replace=item.replace,
        )
        return items


@total_ordering
@dataclass
class DeployResult:
    name: str
    created: int
    deleted: int
    skipped: int
    total: int

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name


class DeployResults(UserList):
    def __init__(self, collection: Iterable[DeployResult], action: Literal["deploy", "clean"], dry_run: bool = False):
        super().__init__(collection)
        self.action = action
        self.dry_run = dry_run

    def create_rich_table(self) -> Table:
        table = Table(title=f"Summary of {self.action} command:")
        prefix = ""
        if self.dry_run:
            prefix = "Would have "
        table.add_column("Resource", justify="right")
        table.add_column(f"{prefix}Created", justify="right", style="green")
        table.add_column(f"{prefix}Deleted", justify="right", style="red")
        table.add_column(f"{prefix}Skipped", justify="right", style="yellow")
        table.add_column("Total", justify="right")
        for item in sorted(self.data):
            table.add_row(
                item.name,
                str(item.created),
                str(item.deleted),
                str(item.skipped),
                str(item.total),
            )

        return table


def deploy_or_clean_resources(
    loader: Loader,
    path: Path,
    ToolGlobals: CDFToolConfig,
    drop: bool = False,
    clean: bool = False,
    action: Literal["deploy", "clean"] = "deploy",
    dry_run: bool = False,
    drop_data: bool = False,
    verbose: bool = False,
) -> DeployResult:
    if action not in ["deploy", "clean"]:
        raise ValueError(f"Invalid action {action}")

    if path.is_file():
        if path.suffix not in loader.filetypes or not loader.filetypes:
            raise ValueError("Invalid file type")
        filepaths = [path]
    elif loader.filetypes:
        filepaths = [file for type_ in loader.filetypes for file in path.glob(f"**/*.{type_}")]
    else:
        filepaths = [file for file in path.glob("**/*")]

    if loader.filename_pattern:
        # This is used by data modelings resources to filter out files that are not of the correct type
        # as these resources share the same folder.
        pattern = re.compile(loader.filename_pattern)
        filepaths = [file for file in filepaths if pattern.match(file.stem)]
    if action == "clean":
        # If we do a clean, we do not want to verify that everything exists wrt data sets, spaces etc.
        items = [loader.load_resource(f, dry_run=True) for f in filepaths]
    else:
        items = [loader.load_resource(f, dry_run) for f in filepaths]
    items = [item for item in items if item is not None]
    nr_of_batches = len(items)
    nr_of_items = sum(len(item) if isinstance(item, Sized) else 1 for item in items)
    if nr_of_items == 0:
        return DeployResult(name=loader.display_name, created=0, deleted=0, skipped=0, total=0)
    if action == "deploy":
        action_word = "Loading" if dry_run else "Uploading"
        print(f"[bold]{action_word} {nr_of_items} {loader.display_name} in {nr_of_batches} batches to CDF...[/]")
    else:
        action_word = "Loading" if dry_run else "Cleaning"
        print(f"[bold]{action_word} {nr_of_items} {loader.display_name} in {nr_of_batches} batches to CDF...[/]")
    batches = [item if isinstance(item, Sized) else [item] for item in items]
    if drop and loader.support_drop and action == "deploy":
        if drop_data and (loader.api_name == "data_modeling.spaces" or loader.api_name == "data_modeling.containers"):
            print(
                f"  --drop-data is specified, will delete existing nodes and edges before before deleting {loader.display_name}."
            )
        else:
            print(f"  --drop is specified, will delete existing {loader.display_name} before uploading.")

    # Deleting resources.
    nr_of_deleted = 0
    if (drop and loader.support_drop) or clean:
        for batch in batches:
            drop_items = [loader.get_id(item) for item in batch]
            if dry_run:
                nr_of_deleted += len(drop_items)
                if verbose:
                    print(f"  Would have deleted {len(drop_items)} {loader.display_name}.")
            else:
                try:
                    nr_of_deleted += loader.delete(drop_items, drop_data)
                except CogniteAPIError as e:
                    if e.code == 404:
                        print(f"  [bold yellow]WARNING:[/] {len(drop_items)} {loader.display_name} do(es) not exist.")
                except CogniteNotFoundError:
                    print(f"  [bold yellow]WARNING:[/] {len(drop_items)} {loader.display_name} do(es) not exist.")
                except Exception as e:
                    print(
                        f"  [bold yellow]WARNING:[/] Failed to delete {len(drop_items)} {loader.display_name}. Error {e}."
                    )
                else:  # Delete succeeded
                    if verbose:
                        print(f"  Deleted {len(drop_items)} {loader.display_name}.")
        if dry_run and action == "clean" and verbose:
            # Only clean command prints this, if not we print it at the end
            print(f"  Would have deleted {nr_of_deleted} {loader.display_name} in total.")

    if action == "clean":
        # Clean Command, only delete.
        return DeployResult(name=loader.display_name, created=0, deleted=nr_of_deleted, skipped=0, total=nr_of_items)

    nr_of_created = 0
    nr_of_skipped = 0
    for batch, filepath in zip(batches, filepaths):
        if not drop and loader.support_upsert:
            if verbose:
                print(f"  Comparing {len(batch)} {loader.display_name} from {filepath}...")
            batch = loader.remove_unchanged(batch)
            if verbose:
                print(f"    {len(batch)} {loader.display_name} to be deployed...")

        if batch:
            if dry_run:
                nr_of_created += len(batch)
            else:
                try:
                    created = loader.create(batch, drop, filepath)
                except Exception as e:
                    print(f"  [bold yellow]WARNING:[/] Failed to upload {loader.display_name}. Error {e}.")
                    ToolGlobals.failed = True
                    return
                else:
                    newly_created = len(created) if created is not None else 0
                    nr_of_created += newly_created
                    nr_of_skipped += len(batch) - newly_created
                    if isinstance(loader, AuthLoader):
                        nr_of_deleted += len(created)
    if verbose:
        prefix = "Would have" if dry_run else ""
        print(
            f"  {prefix} Created {nr_of_created}, Deleted {nr_of_deleted}, Skipped {nr_of_skipped}, Total {nr_of_items}."
        )
    return DeployResult(
        name=loader.display_name,
        created=nr_of_created,
        deleted=nr_of_deleted,
        skipped=nr_of_skipped,
        total=nr_of_items,
    )


LOADER_BY_FOLDER_NAME: dict[str, list[type[Loader]]] = {}
for loader in Loader.__subclasses__():
    if loader.folder_name not in LOADER_BY_FOLDER_NAME:
        LOADER_BY_FOLDER_NAME[loader.folder_name] = []
    LOADER_BY_FOLDER_NAME[loader.folder_name].append(loader)
del loader  # cleanup module namespace



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc
import collections
import inspect
import itertools
import json
import logging
import os
import re
import typing
from collections import UserList
from collections.abc import Collection, Sequence
from dataclasses import dataclass
from functools import total_ordering
from pathlib import Path
from typing import Any, ClassVar, Generic, TypeVar, get_origin

import yaml
from cognite.client import ClientConfig, CogniteClient
from cognite.client.config import global_config
from cognite.client.credentials import OAuthClientCredentials, Token
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError, CogniteAuthError
from cognite.client.utils._text import to_camel_case, to_snake_case
from rich import print

from cognite_toolkit._version import __version__
from cognite_toolkit.cdf_tk._get_type_hints import _TypeHints

logger = logging.getLogger(__name__)


class CDFToolConfig:
    """Configurations for how to store data in CDF

    Properties:
        client: active CogniteClient
    Functions:
        verify_client: verify that the client has correct credentials and specified access capabilities
        verify_dataset: verify that the data set exists and that the client has access to it

    """

    def __init__(self, token: str | None = None, cluster: str | None = None, project: str | None = None) -> None:
        self._data_set_id: int = 0
        self._data_set = None
        self._failed = False
        self._environ = {}
        self._data_set_id_by_external_id: dict[str, id] = {}
        self._existing_spaces: set[str] = set()
        self.oauth_credentials = OAuthClientCredentials(
            token_url="",
            client_id="",
            client_secret="",
            scopes=[],
        )
        # ClientName is used for logging usage of the CDF-Toolkit.
        client_name = f"CDF-Toolkit:{__version__}"

        # CDF_CLUSTER and CDF_PROJECT are minimum requirements and can be overridden
        # when instansiating the class.
        if cluster is not None and len(cluster) > 0:
            self._cluster = cluster
            self._environ["CDF_CLUSTER"] = cluster
        if project is not None and len(project) > 0:
            self._project = project
            self._environ["CDF_PROJECT"] = project
        if token is not None:
            self._environ["CDF_TOKEN"] = token
        if (
            self.environ("CDF_URL", default=None, fail=False) is None
            and self.environ("CDF_CLUSTER", default=None, fail=False) is None
        ):
            # If CDF_URL and CDF_CLUSTER are not set, we may be in a Jupyter notebook in Fusion,
            # and credentials are preset to logged in user (no env vars are set!).
            try:
                self._client = CogniteClient()
            except Exception:
                print(
                    "[bold yellow]WARNING[/] Not able to successfully configure a Cognite client. Requirements: CDF_CLUSTER and CDF_PROJECT environment variables or CDF_TOKEN to a valid OAuth2 token."
                )
            return

        # CDF_CLUSTER and CDF_PROJECT are minimum requirements to know where to connect.
        # Above they were forced default to None and fail was False, here we
        # will fail with an exception if they are not set.
        self._cluster = self.environ("CDF_CLUSTER")
        self._project = self.environ("CDF_PROJECT")
        # CDF_URL is optional, but if set, we use that instead of the default URL using cluster.
        self._cdf_url = self.environ("CDF_URL", f"https://{self._cluster}.cognitedata.com")
        # If CDF_TOKEN is set, we want to use that token instead of client credentials.
        if self.environ("CDF_TOKEN", default=None, fail=False) is not None or token is not None:
            self._client = CogniteClient(
                ClientConfig(
                    client_name=client_name,
                    base_url=self._cdf_url,
                    project=self._project,
                    credentials=Token(token or self.environ("CDF_TOKEN")),
                )
            )
        else:
            # We are now doing OAuth2 client credentials flow, so we need to set the
            # required variables.
            # We can infer scopes and audience from the cluster value.
            # However, the URL to use to retrieve the token, as well as
            # the client id and secret, must be set as environment variables.
            self._scopes = [
                self.environ(
                    "IDP_SCOPES",
                    f"https://{self._cluster}.cognitedata.com/.default",
                )
            ]
            self._audience = self.environ("IDP_AUDIENCE", f"https://{self._cluster}.cognitedata.com")
            self.oauth_credentials = OAuthClientCredentials(
                token_url=self.environ("IDP_TOKEN_URL"),
                client_id=self.environ("IDP_CLIENT_ID"),
                # client secret should not be stored in-code, so we load it from an environment variable
                client_secret=self.environ("IDP_CLIENT_SECRET"),
                scopes=self._scopes,
                audience=self._audience,
            )
            global_config.disable_pypi_version_check = True
            self._client = CogniteClient(
                ClientConfig(
                    client_name=client_name,
                    base_url=self._cdf_url,
                    project=self._project,
                    credentials=self.oauth_credentials,
                )
            )

    def environment_variables(self) -> dict[str, str]:
        return self._environ.copy()

    def as_string(self):
        environment = self._environ.copy()
        if "IDP_CLIENT_SECRET" in environment:
            environment["IDP_CLIENT_SECRET"] = "***"
        if "TRANSFORMATIONS_CLIENT_SECRET" in environment:
            environment["TRANSFORMATIONS_CLIENT_SECRET"] = "***"
        envs = ""
        for e in environment:
            envs += f"  {e}={environment[e]}\n"
        return f"Cluster {self._cluster} with project {self._project} and config:\n{envs}"

    def __str__(self):
        environment = self._environ.copy()
        if "IDP_CLIENT_SECRET" in environment:
            environment["IDP_CLIENT_SECRET"] = "***"
        if "TRANSFORMATIONS_CLIENT_SECRET" in environment:
            environment["TRANSFORMATIONS_CLIENT_SECRET"] = "***"
        return f"Cluster {self._cluster} with project {self._project} and config:\n" + json.dumps(
            environment, indent=2, sort_keys=True
        )

    @property
    # Flag set if something that should have worked failed if a data set is
    # loaded and/or deleted.
    def failed(self) -> bool:
        return self._failed

    @failed.setter
    def failed(self, value: bool):
        self._failed = value

    @property
    def client(self) -> CogniteClient:
        return self._client

    @property
    def project(self) -> str:
        return self._project

    @property
    def data_set_id(self) -> int | None:
        return self._data_set_id if self._data_set_id > 0 else None

    # Use this to ignore the data set when verifying the client's access capabilities
    def clear_dataset(self):
        self._data_set_id = 0
        self._data_set = None

    def environ(self, attr: str, default: str | list[str] | None = None, fail: bool = True) -> str:
        """Helper function to load variables from the environment.

        Use python-dotenv to load environment variables from an .env file before
        using this function.

        If the environment variable has spaces, it will be split into a list of strings.

        Args:
            attr: name of environment variable
            default: default value if environment variable is not set
            fail: if True, raise ValueError if environment variable is not set

        Yields:
            Value of the environment variable
            Raises ValueError if environment variable is not set and fail=True
        """
        if attr in self._environ and self._environ[attr] is not None:
            return self._environ[attr]
        # If the var was none, we want to re-evaluate from environment.
        self._environ[attr] = os.environ.get(attr, None)
        if self._environ[attr] is None:
            if default is None and fail:
                raise ValueError(f"{attr} property is not available as an environment variable and no default set.")
            self._environ[attr] = default
        return self._environ[attr]

    @property
    def data_set(self) -> str:
        return self._data_set

    @data_set.setter
    def data_set(self, value: str):
        if value is None:
            raise ValueError("Please provide an externalId of a dataset.")
        self._data_set = value
        # Since we now have a new configuration, check the dataset and set the id
        self._data_set_id = self.verify_dataset(data_set_external_id=value)

    def verify_client(
        self,
        capabilities: list[dict[str, list[str]]] | None = None,
        data_set_id: int = 0,
        space_id: str | None = None,
    ) -> CogniteClient:
        """Verify that the client has correct credentials and required access rights

        Supply requirement CDF ACLs to verify if you have correct access
        capabilities = {
            "filesAcl": ["READ", "WRITE"],
            "datasetsAcl": ["READ", "WRITE"]
        }
        The data_set_id will be used when verifying that the client has access to the dataset.
        This approach can be reused for any usage of the Cognite Python SDK.

        Args:
            capabilities (dict[list], optional): access capabilities to verify
            data_set_id (int): id of dataset that access should be granted to
            space_id (str): id of space that access should be granted to

        Yields:
            CogniteClient: Verified client with access rights
            Re-raises underlying SDK exception
        """
        capabilities = capabilities or {}
        try:
            # Using the token/inspect endpoint to check if the client has access to the project.
            # The response also includes access rights, which can be used to check if the client has the
            # correct access for what you want to do.
            resp = self.client.iam.token.inspect()
            if resp is None or len(resp.capabilities.data) == 0:
                raise CogniteAuthError("Don't have any access rights. Check credentials.")
        except Exception as e:
            raise e
        scope = {}
        if data_set_id > 0:
            scope["dataSetScope"] = {"ids": [data_set_id]}
        if space_id is not None:
            scope["spaceScope"] = {"ids": [space_id]}
        if space_id is None and data_set_id == 0:
            scope["all"] = {}
        try:
            caps = [
                Capability.load(
                    {
                        cap: {
                            "actions": actions,
                            "scope": scope,
                        },
                    }
                )
                for cap, actions in capabilities.items()
            ] or None
        except Exception:
            raise ValueError(f"Failed to load capabilities from {capabilities}. Wrong syntax?")
        comp = self.client.iam.compare_capabilities(resp.capabilities, caps)
        if len(comp) > 0:
            print(f"Missing necessary CDF access capabilities: {comp}")
            raise CogniteAuthError("Don't have correct access rights.")
        return self._client

    def verify_capabilities(self, capability: Capability | Sequence[Capability]) -> CogniteClient:
        missing_capabilities = self._client.iam.verify_capabilities(capability)
        if len(missing_capabilities) > 0:
            raise CogniteAuthError(f"Missing capabilities: {missing_capabilities}")
        return self._client

    def verify_dataset(self, data_set_external_id: str) -> int:
        """Verify that the configured data set exists and is accessible

        Args:
            data_set_external_id (str): External_id of the data set to verify
        Returns:
            data_set_id (int)
            Re-raises underlying SDK exception
        """
        if data_set_external_id in self._data_set_id_by_external_id:
            return self._data_set_id_by_external_id[data_set_external_id]

        self.verify_client(capabilities={"datasetsAcl": ["READ"]})
        try:
            data_set = self.client.data_sets.retrieve(external_id=data_set_external_id)
        except CogniteAPIError as e:
            raise CogniteAuthError("Don't have correct access rights. Need READ and WRITE on datasetsAcl.") from e
        if data_set is not None:
            self._data_set_id_by_external_id[data_set_external_id] = data_set.id
            return data_set.id
        raise ValueError(
            f"Data set {data_set_external_id} does not exist, you need to create it first. Do this by adding a config file to the data_sets folder."
        )

    def verify_extraction_pipeline(self, external_id: str) -> int:
        """Verify that the configured extraction pipeline exists and is accessible

        Args:
            external_id (str): External id of the extraction pipeline to verify
        Yields:
            extraction pipeline id (int)
            Re-raises underlying SDK exception
        """

        self.verify_client(capabilities={"extractionPipelinesAcl": ["READ"]})
        try:
            pipeline = self.client.extraction_pipelines.retrieve(external_id=external_id)
        except CogniteAPIError as e:
            raise CogniteAuthError("Don't have correct access rights. Need READ on extractionPipelinesAcl.") from e

        if pipeline is not None:
            return pipeline.id
        else:
            print(
                f"  [bold yellow]WARNING[/] Extraction pipeline {external_id} does not exist. It may have been deleted, or not been part of the module."
            )
            return -1

    def verify_spaces(self, space: str | list[str]) -> list[str]:
        """Verify that the configured space exists and is accessible

        Args:
            space (str): External id of the space to verify

        Yields:
            spaces (str)
            Re-raises underlying SDK exception
        """
        if isinstance(space, str):
            space = [space]
        if all([s in self._existing_spaces for s in space]):
            return space

        self.verify_client(capabilities={"dataModelsAcl": ["READ"]})
        try:
            existing = self.client.data_modeling.spaces.retrieve(space)
        except CogniteAPIError as e:
            raise CogniteAuthError("Don't have correct access rights. Need READ on dataModelsAcl.") from e

        if missing := (({space} if isinstance(space, str) else set(space)) - set(existing.as_ids())):
            raise ValueError(
                f"Space {missing} does not exist, you need to create it first. Do this by adding a config file to the data model folder."
            )
        self._existing_spaces.update([space.space for space in existing])
        return [space.space for space in existing]


def load_yaml_inject_variables(filepath: Path, variables: dict[str, str]) -> dict[str, Any] | list[dict[str, Any]]:
    content = filepath.read_text()
    for key, value in variables.items():
        if value is None:
            continue
        content = content.replace("${%s}" % key, value)
    return yaml.safe_load(content)


@dataclass(frozen=True)
class LoadWarning:
    _type: ClassVar[str]
    filepath: Path
    id_value: str
    id_name: str


@total_ordering
@dataclass(frozen=True)
class SnakeCaseWarning(LoadWarning):
    actual: str
    expected: str

    def __lt__(self, other: SnakeCaseWarning) -> bool:
        if not isinstance(other, SnakeCaseWarning):
            return NotImplemented
        return (self.filepath, self.id_value, self.expected, self.actual) < (
            other.filepath,
            other.id_value,
            other.expected,
            other.actual,
        )

    def __eq__(self, other: SnakeCaseWarning) -> bool:
        if not isinstance(other, SnakeCaseWarning):
            return NotImplemented
        return (self.filepath, self.id_value, self.expected, self.actual) == (
            other.filepath,
            other.id_value,
            other.expected,
            other.actual,
        )

    def __str__(self):
        return f"CaseWarning: Got {self.actual!r}. Did you mean {self.expected!r}?"


@total_ordering
@dataclass(frozen=True)
class TemplateVariableWarning(LoadWarning):
    path: str

    def __lt__(self, other: TemplateVariableWarning) -> bool:
        if not isinstance(other, TemplateVariableWarning):
            return NotImplemented
        return (self.id_name, self.id_value, self.path) < (other.id_name, other.id_value, other.path)

    def __eq__(self, other: TemplateVariableWarning) -> bool:
        if not isinstance(other, TemplateVariableWarning):
            return NotImplemented
        return (self.id_name, self.id_value, self.path) == (other.id_name, other.id_value, other.path)

    def __str__(self):
        return f"{type(self).__name__}: Variable {self.id_name!r} has value {self.id_value!r} in file: {self.filepath.name}. Did you forget to change it?"


@total_ordering
@dataclass(frozen=True)
class DataSetMissingWarning(LoadWarning):
    resource_name: str

    def __lt__(self, other: DataSetMissingWarning) -> bool:
        if not isinstance(other, DataSetMissingWarning):
            return NotImplemented
        return (self.id_name, self.id_value, self.filepath) < (other.id_name, other.id_value, other.filepath)

    def __eq__(self, other: DataSetMissingWarning) -> bool:
        if not isinstance(other, DataSetMissingWarning):
            return NotImplemented
        return (self.id_name, self.id_value, self.filepath) == (other.id_name, other.id_value, other.filepath)

    def __str__(self):
        # Avoid circular import
        from cognite_toolkit.cdf_tk.load import TransformationLoader

        if self.filepath.parent.name == TransformationLoader.folder_name:
            return f"{type(self).__name__}: It is recommended to use a data set if source or destination can be scoped with a data set. If not, ignore this warning."
        else:
            return f"{type(self).__name__}: It is recommended that you set dataSetExternalId for {self.resource_name}. This is missing in {self.filepath.name}. Did you forget to add it?"


T_Warning = TypeVar("T_Warning", bound=LoadWarning)


class Warnings(UserList, Generic[T_Warning]):
    def __init__(self, collection: Collection[T_Warning] | None = None):
        super().__init__(collection or [])


class SnakeCaseWarningList(Warnings[SnakeCaseWarning]):
    def __str__(self) -> str:
        output = [""]
        for (file, identifier, id_name), file_warnings in itertools.groupby(
            sorted(self), key=lambda w: (w.filepath, w.id_value, w.id_name)
        ):
            output.append(f"    In File {str(file)!r}")
            output.append(f"    In entry {id_name}={identifier!r}")
            for warning in file_warnings:
                output.append(f"{'    ' * 2}{warning!s}")

        return "\n".join(output)


class TemplateVariableWarningList(Warnings[TemplateVariableWarning]):
    def __str__(self):
        output = [""]
        for path, module_warnings in itertools.groupby(sorted(self), key=lambda w: w.path):
            if path:
                output.append(f"    In Section {str(path)!r}")
            for warning in module_warnings:
                output.append(f"{'    ' * 2}{warning!s}")

        return "\n".join(output)


class DataSetMissingWarningList(Warnings[DataSetMissingWarning]):
    def __str__(self):
        output = [""]
        for filepath, warnings in itertools.groupby(sorted(self), key=lambda w: w.filepath):
            output.append(f"    In file {str(filepath)!r}")
            for warning in warnings:
                output.append(f"{'    ' * 2}{warning!s}")

        return "\n".join(output)


def validate_case_raw(
    raw: dict[str, Any] | list[dict[str, Any]],
    resource_cls: type[CogniteObject],
    filepath: Path,
    identifier_key: str = "externalId",
) -> SnakeCaseWarningList:
    """Checks whether camel casing the raw data would match a parameter in the resource class.

    Args:
        raw: The raw data to check.
        resource_cls: The resource class to check against init method
        filepath: The filepath of the raw data. This is used to pass to the warnings for easy
            grouping of warnings.
        identifier_key: The key to use as identifier. Defaults to "externalId". This is used to pass to the warnings
            for easy grouping of warnings.

    Returns:
        A list of CaseWarning objects.

    """
    return _validate_case_raw(raw, resource_cls, filepath, identifier_key)


def _validate_case_raw(
    raw: dict[str, Any] | list[dict[str, Any]],
    resource_cls: type[CogniteObject],
    filepath: Path,
    identifier_key: str = "externalId",
    identifier_value: str = "",
) -> SnakeCaseWarningList:
    warnings = SnakeCaseWarningList()
    if isinstance(raw, list):
        for item in raw:
            warnings.extend(_validate_case_raw(item, resource_cls, filepath, identifier_key))
        return warnings
    elif not isinstance(raw, dict):
        return warnings

    signature = inspect.signature(resource_cls.__init__)

    is_base_class = inspect.isclass(resource_cls) and any(base is abc.ABC for base in resource_cls.__bases__)
    if is_base_class:
        # If it is a base class, it cannot be instantiated, so it can be any of the
        # subclasses' parameters.
        expected = {
            to_camel_case(parameter)
            for sub in resource_cls.__subclasses__()
            for parameter in inspect.signature(sub.__init__).parameters.keys()
        } - {"self"}
    else:
        expected = set(map(to_camel_case, signature.parameters.keys())) - {"self"}

    actual = set(raw.keys())
    actual_camel_case = set(map(to_camel_case, actual))
    snake_cased = actual - actual_camel_case

    if not identifier_value:
        identifier_value = raw.get(
            identifier_key, raw.get(to_snake_case(identifier_key), f"No identifier {identifier_key}")
        )

    for key in snake_cased:
        if (camel_key := to_camel_case(key)) in expected:
            warnings.append(SnakeCaseWarning(filepath, identifier_value, identifier_key, str(key), str(camel_key)))

    try:
        type_hints_by_name = _TypeHints.get_type_hints_by_name(signature, resource_cls)
    except Exception:
        # If we cannot get type hints, we cannot check if the type is correct.
        return warnings

    for key, value in raw.items():
        if not isinstance(value, dict):
            continue
        if (parameter := signature.parameters.get(to_snake_case(key))) and (
            type_hint := type_hints_by_name.get(parameter.name)
        ):
            if issubclass(type_hint, CogniteObject):
                warnings.extend(_validate_case_raw(value, type_hint, filepath, identifier_key, identifier_value))
                continue

            container_type = get_origin(type_hint)
            if container_type not in [dict, dict, collections.abc.MutableMapping, collections.abc.Mapping]:
                continue
            args = typing.get_args(type_hint)
            if not args:
                continue
            container_key, container_value = args
            if inspect.isclass(container_value) and issubclass(container_value, CogniteObject):
                for sub_key, sub_value in value.items():
                    warnings.extend(
                        _validate_case_raw(sub_value, container_value, filepath, identifier_key, identifier_value)
                    )

    return warnings


def validate_config_yaml(config: dict[str, Any], filepath: Path, path: str = "") -> TemplateVariableWarningList:
    """Checks whether the config file has any issues.

    Currently, this checks for:
        * Non-replaced template variables, such as <change_me>.

    Args:
        config: The config to check.
        filepath: The filepath of the config.yaml.
        path: The path in the config.yaml. This is used recursively by this function.
    """
    warnings = TemplateVariableWarningList()
    pattern = re.compile(r"<.*?>")
    for key, value in config.items():
        if isinstance(value, str) and pattern.match(value):
            warnings.append(TemplateVariableWarning(filepath, value, key, path))
        elif isinstance(value, dict):
            if path:
                path += "."
            warnings.extend(validate_config_yaml(value, filepath, f"{path}{key}"))
    return warnings


def validate_data_set_is_set(
    raw: dict[str, Any] | list[dict[str, Any]],
    resource_cls: type[CogniteObject],
    filepath: Path,
    identifier_key: str = "externalId",
) -> DataSetMissingWarningList:
    warnings = DataSetMissingWarningList()
    signature = inspect.signature(resource_cls.__init__)
    if "data_set_id" not in set(signature.parameters.keys()):
        return warnings

    if isinstance(raw, list):
        for item in raw:
            warnings.extend(validate_data_set_is_set(item, resource_cls, filepath, identifier_key))
        return warnings

    if "dataSetExternalId" in raw or "dataSetId" in raw:
        return warnings

    value = raw.get(identifier_key, raw.get(to_snake_case(identifier_key), f"No identifier {identifier_key}"))
    warnings.append(DataSetMissingWarning(filepath, value, identifier_key, resource_cls.__name__))
    return warnings



from __future__ import annotations

import importlib
import inspect
import typing
from collections import Counter
from typing import Any, get_type_hints

from cognite.client.data_classes._base import CogniteObject


class _TypeHints:
    """
    This class is used to get type hints from the init function of a CogniteObject.

    After Python 3.10, type hints are treated as strings, so we need to evaluate them to get the actual type.
    """

    @classmethod
    def get_type_hints_by_name(cls, signature, resource_cls: type[CogniteObject]) -> dict[str, Any]:
        """
        Get type hints from the init function of a CogniteObject.

        Args:
            signature: The signature of the init function.
            resource_cls: The resource class to get type hints from.
        """
        try:
            type_hint_by_name = get_type_hints(resource_cls.__init__, localns=cls._type_checking)
        except TypeError:
            # Python 3.10 Type hints cannot be evaluated with get_type_hints,
            # ref https://stackoverflow.com/questions/66006087/how-to-use-typing-get-type-hints-with-pep585-in-python3-8
            resource_module_vars = vars(importlib.import_module(resource_cls.__module__))
            resource_module_vars.update(cls._type_checking())
            type_hint_by_name = cls._get_type_hints_3_10(resource_module_vars, signature, vars(resource_cls))
        return type_hint_by_name

    @classmethod
    def _type_checking(cls) -> dict[str, Any]:
        """
        When calling the get_type_hints function, it imports the module with the function TYPE_CHECKING is set to False.

        This function takes all the special types used in data classes and returns them as a dictionary so it
        can be used in the local namespaces.
        """
        import numpy as np
        import numpy.typing as npt
        from cognite.client import CogniteClient

        NumpyDatetime64NSArray = npt.NDArray[np.datetime64]
        NumpyInt64Array = npt.NDArray[np.int64]
        NumpyFloat64Array = npt.NDArray[np.float64]
        NumpyObjArray = npt.NDArray[np.object_]
        return {
            "CogniteClient": CogniteClient,
            "NumpyDatetime64NSArray": NumpyDatetime64NSArray,
            "NumpyInt64Array": NumpyInt64Array,
            "NumpyFloat64Array": NumpyFloat64Array,
            "NumpyObjArray": NumpyObjArray,
        }

    @classmethod
    def _get_type_hints_3_10(
        cls, resource_module_vars: dict[str, Any], signature310: inspect.Signature, local_vars: dict[str, Any]
    ) -> dict[str, Any]:
        return {
            name: cls._create_type_hint_3_10(parameter.annotation, resource_module_vars, local_vars)
            for name, parameter in signature310.parameters.items()
            if name != "self"
        }

    @classmethod
    def _create_type_hint_3_10(
        cls, annotation: str, resource_module_vars: dict[str, Any], local_vars: dict[str, Any]
    ) -> Any:
        if annotation.endswith(" | None"):
            annotation = annotation[:-7]
        try:
            return eval(annotation, resource_module_vars, local_vars)
        except TypeError:
            # Python 3.10 Type Hint
            return cls._type_hint_3_10_to_8(annotation, resource_module_vars, local_vars)

    @classmethod
    def _type_hint_3_10_to_8(
        cls, annotation: str, resource_module_vars: dict[str, Any], local_vars: dict[str, Any]
    ) -> Any:
        if cls._is_vertical_union(annotation):
            alternatives = [
                cls._create_type_hint_3_10(a.strip(), resource_module_vars, local_vars) for a in annotation.split("|")
            ]
            return typing.Union[tuple(alternatives)]
        elif annotation.startswith("dict[") and annotation.endswith("]"):
            if Counter(annotation)[","] > 1:
                key, rest = annotation[5:-1].split(",", 1)
                return dict[key.strip(), cls._create_type_hint_3_10(rest.strip(), resource_module_vars, local_vars)]
            key, value = annotation[5:-1].split(",")
            return dict[
                cls._create_type_hint_3_10(key.strip(), resource_module_vars, local_vars),
                cls._create_type_hint_3_10(value.strip(), resource_module_vars, local_vars),
            ]
        elif annotation.startswith("Mapping[") and annotation.endswith("]"):
            if Counter(annotation)[","] > 1:
                key, rest = annotation[8:-1].split(",", 1)
                return typing.Mapping[
                    key.strip(), cls._create_type_hint_3_10(rest.strip(), resource_module_vars, local_vars)
                ]
            key, value = annotation[8:-1].split(",")
            return typing.Mapping[
                cls._create_type_hint_3_10(key.strip(), resource_module_vars, local_vars),
                cls._create_type_hint_3_10(value.strip(), resource_module_vars, local_vars),
            ]
        elif annotation.startswith("Optional[") and annotation.endswith("]"):
            return typing.Optional[cls._create_type_hint_3_10(annotation[9:-1], resource_module_vars, local_vars)]
        elif annotation.startswith("list[") and annotation.endswith("]"):
            return list[cls._create_type_hint_3_10(annotation[5:-1], resource_module_vars, local_vars)]
        elif annotation.startswith("tuple[") and annotation.endswith("]"):
            return tuple[cls._create_type_hint_3_10(annotation[6:-1], resource_module_vars, local_vars)]
        elif annotation.startswith("typing.Sequence[") and annotation.endswith("]"):
            # This is used in the Sequence data class file to avoid name collision
            return typing.Sequence[cls._create_type_hint_3_10(annotation[16:-1], resource_module_vars, local_vars)]
        raise NotImplementedError(f"Unsupported conversion of type hint {annotation!r}. {cls._error_msg}")

    @classmethod
    def _is_vertical_union(cls, annotation: str) -> bool:
        if "|" not in annotation:
            return False
        parts = [p.strip() for p in annotation.split("|")]
        for part in parts:
            counts = Counter(part)
            if counts["["] != counts["]"]:
                return False
        return True



#!/usr/bin/env python
import shutil
import tempfile
import urllib
import zipfile
from collections.abc import Sequence
from dataclasses import dataclass
from graphlib import TopologicalSorter
from importlib import resources
from pathlib import Path
from typing import Annotated, Optional

import typer
from dotenv import load_dotenv
from rich import print
from rich.panel import Panel

from cognite_toolkit import _version
from cognite_toolkit.cdf_tk import bootstrap
from cognite_toolkit.cdf_tk.load import (
    LOADER_BY_FOLDER_NAME,
    AuthLoader,
    DeployResults,
    deploy_or_clean_resources,
)
from cognite_toolkit.cdf_tk.templates import (
    BUILD_ENVIRONMENT_FILE,
    COGNITE_MODULES,
    CONFIG_FILE,
    CUSTOM_MODULES,
    ENVIRONMENTS_FILE,
    BuildEnvironment,
    build_config,
    generate_config,
    read_yaml_file,
)
from cognite_toolkit.cdf_tk.utils import CDFToolConfig

app = typer.Typer(pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)
auth_app = typer.Typer(
    pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False
)
app.add_typer(auth_app, name="auth")


_AVAILABLE_DATA_TYPES: tuple[str, ...] = tuple(LOADER_BY_FOLDER_NAME)


# Common parameters handled in common callback
@dataclass
class Common:
    override_env: bool
    verbose: bool
    cluster: str
    project: str
    mockToolGlobals: CDFToolConfig


def _version_callback(value: bool):
    if value:
        typer.echo(f"CDF-Toolkit version: {_version.__version__}.")
        raise typer.Exit()


@app.callback(invoke_without_command=True)
def common(
    ctx: typer.Context,
    verbose: Annotated[
        bool,
        typer.Option(
            help="Turn on to get more verbose output when running the commands",
        ),
    ] = False,
    override_env: Annotated[
        bool,
        typer.Option(
            help="Load the .env file in this or the parent directory, but also override currently set environment variables",
        ),
    ] = False,
    cluster: Annotated[
        Optional[str],
        typer.Option(
            envvar="CDF_CLUSTER",
            help="The Cognite Data Fusion cluster to use. Can also be set with the CDF_CLUSTER environment variable.",
        ),
    ] = None,
    project: Annotated[
        Optional[str],
        typer.Option(
            envvar="CDF_PROJECT",
            help="The Cognite Data Fusion project to use. Can also be set with the CDF_PROJECT environment variable.",
        ),
    ] = None,
    version: bool = typer.Option(
        None,
        "--version",
        help="See which version of the tooklit and the templates are installed.",
        callback=_version_callback,
    ),
):
    """The cdf-tk tool is used to build and deploy Cognite Data Fusion project configurations from the command line or through CI/CD pipelines.

    Each of the main commands has a separate help, e.g. `cdf-tk build --help` or `cdf-tk deploy --help`.

    You can find the documation at https://developer.cognite.com/sdks/toolkit/
    and the template reference documentation at https://developer.cognite.com/sdks/toolkit/references/configs
    """
    if ctx.invoked_subcommand is None:
        print(
            "[bold]A tool to manage and deploy Cognite Data Fusion project configurations from the command line or through CI/CD pipelines.[/]"
        )
        print("[bold yellow]Usage:[/] cdf-tk [OPTIONS] COMMAND [ARGS]...")
        print("       Use --help for more information.")
        return
    if override_env:
        print("  [bold yellow]WARNING:[/] Overriding environment variables with values from .env file...")
        if cluster is not None or project is not None:
            print("            --cluster or --project is set and will override .env file values.")
    if not (Path.cwd() / ".env").is_file():
        if not (Path.cwd().parent / ".env").is_file():
            print("[bold yellow]WARNING:[/] No .env file found in current or parent directory.")
        else:
            if verbose:
                print("Loading .env file found in parent directory.")
            load_dotenv("../.env", override=override_env)
    else:
        if verbose:
            print("Loading .env file found in current directory.")
        load_dotenv(".env", override=override_env)
    ctx.obj = Common(
        verbose=verbose,
        override_env=override_env,
        cluster=cluster,
        project=project,
        mockToolGlobals=None,
    )


@app.command("build")
def build(
    ctx: typer.Context,
    source_dir: Annotated[
        Optional[str],
        typer.Argument(
            help="Where to find the module templates to build from",
            allow_dash=True,
        ),
    ] = "./",
    build_dir: Annotated[
        Optional[str],
        typer.Option(
            "--build-dir",
            "-b",
            help="Where to save the built module files",
        ),
    ] = "./build",
    build_env: Annotated[
        Optional[str],
        typer.Option(
            "--env",
            "-e",
            help="Build environment to build for",
        ),
    ] = "dev",
    clean: Annotated[
        Optional[bool],
        typer.Option(
            "--clean",
            "-c",
            help="Delete the build directory before building the configurations",
        ),
    ] = False,
) -> None:
    """Build configuration files from the module templates to a local build directory."""
    source_dir = Path(source_dir)
    if not source_dir.is_dir():
        print(f"  [bold red]ERROR:[/] {source_dir} does not exist")
        exit(1)
    environment_file = Path.cwd() / ENVIRONMENTS_FILE
    if not environment_file.is_file() and not (environment_file := source_dir / ENVIRONMENTS_FILE).is_file():
        print(f"  [bold red]ERROR:[/] {environment_file} does not exist")
        exit(1)
    config_file = Path.cwd() / CONFIG_FILE
    if not config_file.is_file() and not (config_file := source_dir / CONFIG_FILE).is_file():
        print(f"  [bold red]ERROR:[/] {config_file} does not exist")
        exit(1)
    print(
        Panel(
            f"[bold]Building config files from templates into {build_dir!s} for environment {build_env} using {source_dir!s} as sources...[/bold]"
            f"\n[bold]Environment file:[/] {environment_file.absolute().relative_to(Path.cwd())!s} and [bold]config file:[/] {config_file.absolute().relative_to(Path.cwd())!s}"
        )
    )
    print(f"  Environment is {build_env}, using that section in {ENVIRONMENTS_FILE!s}.\n")
    build_ = BuildEnvironment.load(read_yaml_file(environment_file), build_env, "build")
    build_.set_environment_variables()

    build_config(
        build_dir=Path(build_dir),
        source_dir=source_dir,
        config_file=config_file,
        build=build_,
        clean=clean,
        verbose=ctx.obj.verbose,
    )


@app.command("deploy")
def deploy(
    ctx: typer.Context,
    build_dir: Annotated[
        Optional[str],
        typer.Argument(
            help="Where to find the module templates to deploy from. Defaults to current directory.",
            allow_dash=True,
        ),
    ] = "./build",
    build_env: Annotated[
        Optional[str],
        typer.Option(
            "--env",
            "-e",
            help="CDF project environment to build for. Defined in environments.yaml. Defaults to dev.",
        ),
    ] = "dev",
    interactive: Annotated[
        Optional[bool],
        typer.Option(
            "--interactive",
            "-i",
            help="Whether to use interactive mode when deciding which modules to deploy.",
        ),
    ] = False,
    drop: Annotated[
        Optional[bool],
        typer.Option(
            "--drop",
            "-d",
            help="Whether to drop existing configurations, drop per resource if present.",
        ),
    ] = False,
    drop_data: Annotated[
        Optional[bool],
        typer.Option(
            "--drop-data",
            help="Whether to drop existing data in data model containers and spaces.",
        ),
    ] = False,
    dry_run: Annotated[
        Optional[bool],
        typer.Option(
            "--dry-run",
            "-r",
            help="Whether to do a dry-run, do dry-run if present.",
        ),
    ] = False,
    include: Annotated[
        Optional[list[str]],
        typer.Option(
            "--include",
            "-i",
            help=f"Specify which resources to deploy, available options: {_AVAILABLE_DATA_TYPES}.",
        ),
    ] = None,
) -> None:
    """Deploy one or more resource types from the built configurations to a CDF project environment of your choice (as set in environments.yaml)."""
    # Override cluster and project from the options/env variables
    if ctx.obj.mockToolGlobals is not None:
        ToolGlobals = ctx.obj.mockToolGlobals
    else:
        ToolGlobals = CDFToolConfig(cluster=ctx.obj.cluster, project=ctx.obj.project)

    build_ = BuildEnvironment.load(read_yaml_file(Path(build_dir) / BUILD_ENVIRONMENT_FILE), build_env, "deploy")
    build_.set_environment_variables()

    print(Panel(f"[bold]Deploying config files from {build_dir} to environment {build_env}...[/]"))
    build_path = Path(build_dir)
    if not build_path.is_dir():
        typer.echo(
            f"  [bold yellow]WARNING:[/] {build_dir} does not exists. Did you forget to run `cdf-tk build` first?"
        )
        exit(1)

    include = _process_include(include, interactive)
    print(ToolGlobals.as_string())

    # The 'auth' loader is excluded, as it is run twice,
    # once with all_scoped_skipped_validation and once with resource_scoped_only
    selected_loaders = {
        LoaderCls: LoaderCls.dependencies
        for folder_name, loader_classes in LOADER_BY_FOLDER_NAME.items()
        if folder_name in include and folder_name != "auth" and (build_path / folder_name).is_dir()
        for LoaderCls in loader_classes
    }

    arguments = dict(
        ToolGlobals=ToolGlobals,
        drop=drop,
        action="deploy",
        dry_run=dry_run,
        drop_data=drop_data,
        verbose=ctx.obj.verbose,
    )
    results = DeployResults([], "deploy", dry_run=dry_run)
    if "auth" in include and (directory := (Path(build_dir) / "auth")).is_dir():
        # First, we need to get all the generic access, so we can create the rest of the resources.
        print("[bold]EVALUATING auth resources (groups) with ALL scope...[/]")
        result = deploy_or_clean_resources(
            AuthLoader.create_loader(ToolGlobals, target_scopes="all_scoped_skipped_validation"),
            directory,
            **arguments,
        )
        results.append(result)
        if ToolGlobals.failed:
            print("[bold red]ERROR: [/] Failure to deploy auth (groups) with ALL scope as expected.")
            exit(1)
    for LoaderCls in TopologicalSorter(selected_loaders).static_order():
        result = deploy_or_clean_resources(
            LoaderCls.create_loader(ToolGlobals),
            build_path / LoaderCls.folder_name,
            **arguments,
        )
        results.append(result)
        if ToolGlobals.failed:
            if results:
                print(results.create_rich_table())
            print(f"[bold red]ERROR: [/] Failure to load {LoaderCls.folder_name} as expected.")
            exit(1)

    if "auth" in include and (directory := (Path(build_dir) / "auth")).is_dir():
        # Last, we need to get all the scoped access, as the resources should now have been created.
        print("[bold]EVALUATING auth resources scoped to resources...[/]")
        result = deploy_or_clean_resources(
            AuthLoader.create_loader(ToolGlobals, target_scopes="resource_scoped_only"),
            directory,
            **arguments,
        )
        results.append(result)
    print(results.create_rich_table())
    if ToolGlobals.failed:
        print("[bold red]ERROR: [/] Failure to deploy auth (groups) scoped to resources as expected.")
        exit(1)


@app.command("clean")
def clean(
    ctx: typer.Context,
    build_dir: Annotated[
        Optional[str],
        typer.Argument(
            help="Where to find the module templates to clean from. Defaults to ./build directory.",
            allow_dash=True,
        ),
    ] = "./build",
    build_env: Annotated[
        Optional[str],
        typer.Option(
            "--env",
            "-e",
            help="CDF project environment to use for cleaning.",
        ),
    ] = "dev",
    interactive: Annotated[
        Optional[bool],
        typer.Option(
            "--interactive",
            "-i",
            help="Whether to use interactive mode when deciding which resource types to clean.",
        ),
    ] = False,
    dry_run: Annotated[
        Optional[bool],
        typer.Option(
            "--dry-run",
            "-r",
            help="Whether to do a dry-run, do dry-run if present",
        ),
    ] = False,
    include: Annotated[
        Optional[list[str]],
        typer.Option(
            "--include",
            "-i",
            help=f"Specify which resource types to deploy, supported types: {_AVAILABLE_DATA_TYPES}",
        ),
    ] = None,
) -> None:
    """Clean up a CDF environment as set in environments.yaml restricted to the entities in the configuration files in the build directory."""
    # Override cluster and project from the options/env variables
    if ctx.obj.mockToolGlobals is not None:
        ToolGlobals = ctx.obj.mockToolGlobals
    else:
        ToolGlobals = CDFToolConfig(cluster=ctx.obj.cluster, project=ctx.obj.project)

    build_ = BuildEnvironment.load(read_yaml_file(Path(build_dir) / BUILD_ENVIRONMENT_FILE), build_env, "clean")
    build_.set_environment_variables()

    Panel(f"[bold]Cleaning environment {build_env} based on config files from {build_dir}...[/]")
    build_path = Path(build_dir)
    if not build_path.is_dir():
        typer.echo(
            f"  [bold yellow]WARNING:[/] {build_dir} does not exists. Did you forget to run `cdf-tk build` first?"
        )
        exit(1)

    include = _process_include(include, interactive)
    print(ToolGlobals.as_string())

    # The 'auth' loader is excluded, as it is run at the end.
    selected_loaders = {
        LoaderCls: LoaderCls.dependencies
        for folder_name, loader_classes in LOADER_BY_FOLDER_NAME.items()
        if folder_name in include and folder_name != "auth" and (build_path / folder_name).is_dir()
        for LoaderCls in loader_classes
    }

    print(ToolGlobals.as_string())
    if ToolGlobals.failed:
        print("[bold red]ERROR: [/] Failure to delete data models as expected.")
        exit(1)
    results = DeployResults([], "clean", dry_run=dry_run)
    for LoaderCls in reversed(list(TopologicalSorter(selected_loaders).static_order())):
        result = deploy_or_clean_resources(
            LoaderCls.create_loader(ToolGlobals),
            build_path / LoaderCls.folder_name,
            ToolGlobals,
            drop=True,
            action="clean",
            drop_data=True,
            dry_run=dry_run,
            verbose=ctx.obj.verbose,
        )
        results.append(result)
        if ToolGlobals.failed:
            if results:
                print(results.create_rich_table())
            print(f"[bold red]ERROR: [/] Failure to clean {LoaderCls.folder_name} as expected.")
            exit(1)
    if "auth" in include and (directory := (Path(build_dir) / "auth")).is_dir():
        result = deploy_or_clean_resources(
            AuthLoader.create_loader(ToolGlobals, target_scopes="all"),
            directory,
            ToolGlobals,
            drop=True,
            clean=True,
            action="clean",
            dry_run=dry_run,
            verbose=ctx.obj.verbose,
        )
        results.append(result)
    print(results.create_rich_table())
    if ToolGlobals.failed:
        print("[bold red]ERROR: [/] Failure to clean auth as expected.")
        exit(1)


@auth_app.callback(invoke_without_command=True)
def auth_main(ctx: typer.Context):
    """Test, validate, and configure authentication and authorization for CDF projects."""
    if ctx.invoked_subcommand is None:
        print("Use [bold yellow]cdf-tk auth --help[/] for more information.")


@auth_app.command("verify")
def auth_verify(
    ctx: typer.Context,
    dry_run: Annotated[
        Optional[bool],
        typer.Option(
            "--dry-run",
            "-r",
            help="Whether to do a dry-run, do dry-run if present.",
        ),
    ] = False,
    interactive: Annotated[
        Optional[bool],
        typer.Option(
            "--interactive",
            "-i",
            help="Will run the verification in interactive mode, prompting for input. Used to bootstrap a new project.",
        ),
    ] = False,
    group_file: Annotated[
        Optional[str],
        typer.Option(
            "--group-file",
            "-f",
            help="Path to group yaml configuration file to use for group verification. Defaults to readwrite.all.group.yaml from the cdf_auth_readwrite_all common module.",
        ),
    ] = f"/{COGNITE_MODULES}/common/cdf_auth_readwrite_all/auth/readwrite.all.group.yaml",
    update_group: Annotated[
        Optional[int],
        typer.Option(
            "--update-group",
            "-u",
            help="Used to update an existing group with the configurations from the configuration file. Set to the group id to update or 1 to update the default write-all group (if the tool is only member of one group).",
        ),
    ] = 0,
    create_group: Annotated[
        Optional[str],
        typer.Option(
            "--create-group",
            "-c",
            help="Used to create a new group with the configurations from the configuration file. Set to the source id that the new group should be configured with.",
        ),
    ] = None,
):
    """When you have the necessary information about your identity provider configuration,
    you can use this command to configure the tool and verify that the token has the correct access rights to the project.
    It can also create a group with the correct access rights, defaulting to write-all group
    meant for an admin/CICD pipeline.

    As a minimum, you need the CDF project name, the CDF cluster, an identity provider token URL, and a service account client ID
    and client secret (or an OAuth2 token set in CDF_TOKEN environment variable).

    Needed capabilites for bootstrapping:
    "projectsAcl": ["LIST", "READ"],
    "groupsAcl": ["LIST", "READ", "CREATE", "UPDATE", "DELETE"]

    The default bootstrap group configuration is readwrite.all.group.yaml from the cdf_auth_readwrite_all common module.
    """
    if create_group is not None and update_group != 0:
        print("[bold red]ERROR: [/] --create-group and --update-group are mutually exclusive.")
        exit(1)
    if ctx.obj.mockToolGlobals is not None:
        ToolGlobals = ctx.obj.mockToolGlobals
    else:
        ToolGlobals = CDFToolConfig(cluster=ctx.obj.cluster, project=ctx.obj.project)
    bootstrap.check_auth(
        ToolGlobals,
        group_file=group_file,
        update_group=update_group,
        create_group=create_group,
        interactive=interactive,
        dry_run=dry_run,
        verbose=ctx.obj.verbose,
    )
    if ToolGlobals.failed:
        print("[bold red]ERROR: [/] Failure to verify access rights.")
        exit(1)


@app.command("init")
def main_init(
    ctx: typer.Context,
    dry_run: Annotated[
        Optional[bool],
        typer.Option(
            "--dry-run",
            "-r",
            help="Whether to do a dry-run, do dry-run if present.",
        ),
    ] = False,
    upgrade: Annotated[
        Optional[bool],
        typer.Option(
            "--upgrade",
            "-u",
            help="Will upgrade templates in place without overwriting existing config.yaml and other files.",
        ),
    ] = False,
    git: Annotated[
        Optional[str],
        typer.Option(
            "--git",
            "-g",
            help="Will download the latest templates from the git repository branch specified. Use `main` to get the very latest templates.",
        ),
    ] = None,
    no_backup: Annotated[
        Optional[bool],
        typer.Option(
            "--no-backup",
            help="Will skip making a backup before upgrading.",
        ),
    ] = False,
    clean: Annotated[
        Optional[bool],
        typer.Option(
            "--clean",
            help="Will delete the new_project directory before starting.",
        ),
    ] = False,
    init_dir: Annotated[
        Optional[str],
        typer.Argument(
            help="Directory path to project to initialize or upgrade with templates.",
        ),
    ] = "new_project",
):
    """Initialize or upgrade a new CDF project with templates."""

    files_to_copy = []
    dirs_to_copy = []
    if not upgrade:
        files_to_copy.extend(
            [
                "environments.yaml",
                "README.md",
                ".gitignore",
                ".env.tmpl",
            ]
        )
        dirs_to_copy.append(CUSTOM_MODULES)
    module_dirs_to_copy = [
        COGNITE_MODULES,
    ]
    template_dir = resources.files("cognite_toolkit")
    target_dir = Path.cwd() / f"{init_dir}"
    if target_dir.exists():
        if not upgrade:
            if clean:
                if dry_run:
                    print(f"Would clean out directory {target_dir}...")
                else:
                    print(f"Cleaning out directory {target_dir}...")
                    shutil.rmtree(target_dir)
            else:
                print(f"Directory {target_dir} already exists.")
                exit(1)
        else:
            print(f"[bold]Upgrading directory {target_dir}...[/b]")
    elif upgrade:
        print(f"Found no directory {target_dir} to upgrade.")
        exit(1)
    if not dry_run and not upgrade:
        target_dir.mkdir(exist_ok=True)
    if upgrade:
        print("  Will upgrade modules and files in place, config.yaml files will not be touched.")
    print(f"Will copy these files to {target_dir}:")
    print(files_to_copy)
    print(f"Will copy these module directories to {target_dir}:")
    print(module_dirs_to_copy)
    print(f"Will copy these directories to {target_dir}:")
    print(dirs_to_copy)
    extract_dir = None
    if upgrade and git is not None:
        toolkit_github_url = f"https://github.com/cognitedata/cdf-project-templates/archive/refs/heads/{git}.zip"
        extract_dir = tempfile.mkdtemp(prefix="git.", suffix=".tmp", dir=Path.cwd())
        print(f"Upgrading templates from https://github.com/cognitedata/cdf-project-templates, branch {git}...")
        print(
            "  [bold yellow]WARNING:[/] You are only upgrading templates, not the cdf-tk tool. Your current version may not support the new templates."
        )
        if not dry_run:
            try:
                zip_path, _ = urllib.request.urlretrieve(toolkit_github_url)
                with zipfile.ZipFile(zip_path, "r") as f:
                    f.extractall(extract_dir)
            except Exception:
                print(
                    f"Failed to download templates. Are you sure that the branch {git} exists in"
                    + "the https://github.com/cognitedata/cdf-project-templatesrepository?\n{e}"
                )
                exit(1)
        template_dir = Path(extract_dir) / f"cdf-project-templates-{git}" / "cognite_toolkit"
    for f in files_to_copy:
        if dry_run and ctx.obj.verbose:
            print("Would copy file", f, "to", target_dir)
        elif not dry_run:
            if ctx.obj.verbose:
                print("Copying file", f, "to", target_dir)
            shutil.copyfile(Path(template_dir) / f, target_dir / f)
    for d in dirs_to_copy:
        if dry_run and ctx.obj.verbose:
            if upgrade:
                print("Would copy and overwrite directory", d, "to", target_dir)
            else:
                print("Would copy directory", d, "to", target_dir)
        elif not dry_run:
            if ctx.obj.verbose:
                print("Copying directory", d, "to", target_dir)
            shutil.copytree(Path(template_dir) / d, target_dir / d, dirs_exist_ok=True)
    if upgrade and not no_backup:
        if dry_run:
            if ctx.obj.verbose:
                print(f"Would have backed up {target_dir}")
        else:
            backup_dir = tempfile.mkdtemp(prefix=f"{target_dir.name}.", suffix=".bck", dir=Path.cwd())
            if ctx.obj.verbose:
                print(f"Backing up {target_dir} to {backup_dir}...")
            shutil.copytree(Path(target_dir), Path(backup_dir), dirs_exist_ok=True)
    elif upgrade:
        print("[bold yellow]WARNING:[/] --no-backup is specified, no backup will be made.")
    for d in module_dirs_to_copy:
        if not dry_run:
            (Path(target_dir) / d).mkdir(exist_ok=True)
        if ctx.obj.verbose:
            if dry_run:
                print(f"Would have copied modules in {d}")
            else:
                print(f"Copying modules in {d}...")
        if not dry_run:
            shutil.copytree(Path(template_dir / d), target_dir / d, dirs_exist_ok=True)
    if extract_dir is not None:
        shutil.rmtree(extract_dir)
    if not dry_run:
        if upgrade:
            print(f"You project in {target_dir} was upgraded.")
        else:
            print(f"A new project was created in {target_dir}.")
        if upgrade:
            print("  All default variables from the modules have been upgraded.")
            print("  Please check you config.yaml file for new default variables that may need to be changed.")

    config_filepath = target_dir / "config.yaml"
    if not dry_run:
        if clean or not config_filepath.exists():
            config_str, _ = generate_config(target_dir)
            config_filepath.write_text(config_str)
            print(f"Created your config.yaml file in {target_dir}.")
        else:
            current = config_filepath.read_text()
            config_str, difference = generate_config(target_dir, existing_config=current)
            config_filepath.write_text(config_str)
            print(str(difference))


def _process_include(include: Optional[list[str]], interactive: bool) -> list[str]:
    if include and (invalid_types := set(include).difference(_AVAILABLE_DATA_TYPES)):
        print(
            f"  [bold red]ERROR:[/] Invalid resource types specified: {invalid_types}, available types: {_AVAILABLE_DATA_TYPES}"
        )
        exit(1)
    include = include or list(_AVAILABLE_DATA_TYPES)
    if interactive:
        include = _select_data_types(include)
    return include


def _select_data_types(include: Sequence[str]) -> list[str]:
    mapping: dict[int, str] = {}
    for i, datatype in enumerate(include):
        print(f"[bold]{i})[/] {datatype}")
        mapping[i] = datatype
    print("\na) All")
    print("q) Quit")
    answer = input("Select resource types to include: ")
    if answer.casefold() == "a":
        return list(include)
    elif answer.casefold() == "q":
        exit(0)
    else:
        try:
            return [mapping[int(answer)]]
        except ValueError:
            print(f"Invalid selection: {answer}")
            exit(1)


if __name__ == "__main__":
    app()
